# ATTRIBUTION: this file originates from https://github.com/hfawaz/dl-4-tsc
# It was published on github.com by Ismail Fawaz, Hassan and Forestier, Germain and Weber, Jonathan and Idoumghar, Lhassane and Muller, Pierre-Alain
# for their work "Deep learning for time series classification: a review", in Data Mining and Knowledge Discovery.
# LICENSE is GNU GENERAL PUBLIC LICENSE version 3 (29 June 2007).
# Adopted by BLINDFORREVIEW for 'useckit'.
# Please cite:
"""
@article{IsmailFawaz2018deep,
  Title                    = {Deep learning for time series classification: a review},
  Author                   = {Ismail Fawaz, Hassan and Forestier, Germain and Weber, Jonathan and Idoumghar, Lhassane and Muller, Pierre-Alain},
  journal                  = {Data Mining and Knowledge Discovery},
  Year                     = {2019},
  volume                   = {33},
  number                   = {4},
  pages                    = {917--963},
}
"""
import gc
import os
import pickle
import time

# model twi-esn
import numpy as np
from scipy import sparse
from scipy.sparse import linalg as slinalg
from sklearn.linear_model import Ridge
from sklearn.model_selection import train_test_split

from .tsc_prediction_model_base import TSCBasePredictionModel
from ....util.dataset import Dataset
from ....util.utils import calculate_metrics, create_directory


class dl4tsc_twiesn(TSCBasePredictionModel):
    def __init__(self, output_dir: str = "TWIESN_model_out"):
        super().__init__(output_dir=output_dir)
        # hyperparameters
        first_config = {'N_x': 250, 'connect': 0.5, 'scaleW_in': 1.0, 'lamda': 0.0}
        second_config = {'N_x': 250, 'connect': 0.5, 'scaleW_in': 2.0, 'lamda': 0.05}
        third_config = {'N_x': 500, 'connect': 0.1, 'scaleW_in': 2.0, 'lamda': 0.05}
        fourth_config = {'N_x': 800, 'connect': 0.1, 'scaleW_in': 2.0, 'lamda': 0.05}
        self.configs = [first_config, second_config, third_config, fourth_config]
        self.rho_s = [0.55, 0.9, 2.0, 5.0]
        self.alpha = 0.1  # leaky rate
        self.y_pred = None
        self.y_pred_reshaped = None

    def persist_model_trainables(self, path_like: str = 'saved_trainable_values', **kwargs):
        raise NotImplementedError()

    def restore_model_trainables(self, path_like, dataset: Dataset = None, **kwargs):
        raise NotImplementedError()

    def init_matrices(self):
        self.W_in = (2.0 * np.random.rand(self.N_x, self.num_dim) - 1.0) / (2.0 * self.scaleW_in)

        converged = False

        i = 0

        # repeat because could not converge to find eigenvalues
        while (not converged):
            i += 1

            # generate sparse, uniformly distributed weights
            self.W = sparse.rand(self.N_x, self.N_x, density=self.connect).todense()

            # ensure that the non-zero values are uniformly distributed
            self.W[np.where(self.W > 0)] -= 0.5

            try:
                # get the largest eigenvalue
                eig, _ = slinalg.eigs(self.W, k=1, which='LM')
                converged = True
            except:
                print('not converged ', i)
                continue

        # adjust the spectral radius
        self.W /= np.abs(eig) / self.rho

    def compute_state_matrix(self, x_in):
        # number of instances
        n = x_in.shape[0]
        # the state matrix to be computed
        X_t = np.zeros((n, self.T, self.N_x), dtype=np.float64)
        # previous state matrix
        X_t_1 = np.zeros((n, self.N_x), dtype=np.float64)
        # loop through each time step
        for t in range(self.T):
            # get all the time series data points for the time step t
            curr_in = x_in[:, t, :]
            # calculate the linear activation
            curr_state = np.tanh(self.W_in.dot(curr_in.T) + self.W.dot(X_t_1.T)).T
            # apply leakage
            curr_state = (1 - self.alpha) * X_t_1 + self.alpha * curr_state
            # save in previous state
            X_t_1 = curr_state
            # save in state matrix
            X_t[:, t, :] = curr_state

        return X_t

    def reshape_prediction(self, y_pred, num_instances, length_series):
        # reshape so the first axis has the number of instances
        new_y_pred = y_pred.reshape(num_instances, length_series, y_pred.shape[-1])
        # average the predictions of instances
        new_y_pred = np.average(new_y_pred, axis=1)
        # get the label with maximum prediction over the last label axis
        new_y_pred = np.argmax(new_y_pred, axis=1)
        return new_y_pred

    def train(self):
        start_time = time.time()

        ################
        ### Training ###
        ################

        # init the matrices
        self.init_matrices()
        # compute the state matrices which is the new feature space
        state_matrix = self.compute_state_matrix(self.x_train)
        # add the input to form the new feature space and transform to
        # the new feature space to be feeded to the classifier
        new_x_train = np.concatenate((self.x_train, state_matrix), axis=2).reshape(
            self.N * self.T, self.num_dim + self.N_x)
        # memory free
        state_matrix = None
        gc.collect()
        # transform the corresponding labels
        new_labels = np.repeat(self.y_train, self.T, axis=0)
        # new model
        ridge_classifier = Ridge(alpha=self.lamda)
        # fit the new feature space
        ridge_classifier.fit(new_x_train, new_labels)

        ################
        ## Validation ##
        ################
        # compute state matrix for validation set
        state_matrix = self.compute_state_matrix(self.x_val)
        # add the input to form the new feature space and transform to
        # the new feature space to be feeded to the classifier
        new_x_val = np.concatenate((self.x_val, state_matrix), axis=2).reshape(
            self.x_val.shape[0] * self.T, self.num_dim + self.N_x)
        # get the prediction on the train set
        y_pred_val = ridge_classifier.predict(new_x_val)
        # reconstruct the training prediction
        y_pred_val = self.reshape_prediction(y_pred_val, self.x_val.shape[0], self.T)
        # get the metrics for the train
        df_val_metrics = calculate_metrics(np.argmax(self.y_val, axis=1), y_pred_val, 0.0)
        # get the train accuracy
        train_acc = df_val_metrics['accuracy'][0]

        ###############
        ### Testing ###
        ###############

        # get the predicition on the test set
        # transform the test set to the new features
        state_matrix = self.compute_state_matrix(self.x_test)
        # add the input to form the new feature space and transform to the new feature space to be feeded to the classifier
        new_x_test = np.concatenate((self.x_test, state_matrix), axis=2).reshape(self.x_test.shape[0] * self.T,
                                                                                 self.num_dim + self.N_x)
        # memory free
        state_matrix = None
        gc.collect()
        # get the prediction on the test set
        self.y_pred = y_pred = ridge_classifier.predict(new_x_test)
        # reconstruct the test predictions
        self.y_pred_reshaped = y_pred_reshaped = self.reshape_prediction(y_pred, self.x_test.shape[0], self.T)

        duration = time.time() - start_time
        # get the metrics for the test predictions
        df_metrics = calculate_metrics(self.y_true, y_pred_reshaped, duration)

        # get the output layer weights
        self.W_out = ridge_classifier.coef_
        gc.collect()
        # save the model
        np.savetxt(os.path.join(self.output_dir, 'W_in.txt'), self.W_in)
        np.savetxt(os.path.join(self.output_dir, 'W.txt'), self.W)
        np.savetxt(os.path.join(self.output_dir, 'W_out.txt'), self.W_out)

        # save the metrics
        df_metrics.to_csv(os.path.join(self.output_dir, 'df_metrics.csv'), index=False)

        # return the training accuracy and the prediction metrics on the test set
        return df_metrics, train_acc, ridge_classifier

    def fit(self, dataset: Dataset, *args, **kwargs):
        self.x_train, self.x_val, self.y_train, self.y_val, y_true, input_shape, nb_classes = self.convert_dataset_to_legacy_values(dataset)

        best_train_acc = -1

        self.num_dim = self.x_train.shape[2]
        self.T = self.x_train.shape[1]
        self.x_test = self.x_val
        self.y_true = y_true
        self.y_test = self.y_val

        self.N = self.x_train.shape[0]

        # limit the hyperparameter search if dataset is too big
        if self.x_train.shape[0] > 1000 or self.x_test.shape[0] > 1000:
            for config in self.configs:
                config['N_x'] = 100
            self.configs = [self.configs[0], self.configs[1], self.configs[2]]

        output_directory_root = self.output_dir
        # grid search
        for idx_config in range(len(self.configs)):
            for rho in self.rho_s:
                self.rho = rho
                self.N_x = self.configs[idx_config]['N_x']
                self.connect = self.configs[idx_config]['connect']
                self.scaleW_in = self.configs[idx_config]['scaleW_in']
                self.lamda = self.configs[idx_config]['lamda']
                self.output_dir = os.path.join(output_directory_root,
                                               'hyper_param_search',
                                               'config_' + str(idx_config),
                                               'rho_' + str(rho))
                create_directory(self.output_dir)
                df_metrics, train_acc, clf = self.train()

                if best_train_acc < train_acc:
                    best_train_acc = train_acc
                    self.y_pred_best = self.y_pred
                    self.y_pred_reshaped_best = self.y_pred_reshaped
                    df_metrics.to_csv(os.path.join(output_directory_root, 'df_metrics.csv'), index=False)
                    np.savetxt(os.path.join(output_directory_root, 'W_in.txt'), self.W_in)
                    np.savetxt(os.path.join(output_directory_root, 'W.txt'), self.W)
                    np.savetxt(os.path.join(output_directory_root, 'W_out.txt'), self.W_out)
                    with open(os.path.join(output_directory_root, 'ridge.pkl'), 'wb') as file:
                        pickle.dump(clf, file)
                    with open(os.path.join(output_directory_root, 'T.pkl'), 'wb') as file:
                        pickle.dump(self.T, file)
                    with open(os.path.join(output_directory_root, 'N_x.pkl'), 'wb') as file:
                        pickle.dump(self.N_x, file)
                    with open(os.path.join(output_directory_root, 'alpha.pkl'), 'wb') as file:
                        pickle.dump(self.alpha, file)

                gc.collect()

    def predict(self, x_test, y_true, x_train, y_train, y_test, return_df_metrics=True):
        from pathlib import Path
        modelroot = Path(self.output_dir).parent.parent.parent
        with open(modelroot.joinpath("ridge.pkl"), 'rb') as file:
            ridge_classifier = pickle.load(file)
        with open(modelroot.joinpath("T.pkl"), 'rb') as file:
            self.T = pickle.load(file)
        with open(modelroot.joinpath("N_x.pkl"), 'rb') as file:
            self.N_x = pickle.load(file)
        with open(modelroot.joinpath("alpha.pkl"), 'rb') as file:
            self.alpha = pickle.load(file)
        self.W_in = np.loadtxt(modelroot.joinpath('W_in.txt'))
        self.W = np.loadtxt(modelroot.joinpath('W.txt'))
        self.W_out = np.loadtxt(modelroot.joinpath('W_out.txt'))

        # get the predicition on the test set
        # transform the test set to the new features
        state_matrix = self.compute_state_matrix(x_test)
        # add the input to form the new feature space and transform to the new feature space to be feeded to the classifier
        new_x_test = np.concatenate((x_test, state_matrix), axis=2).reshape(x_test.shape[0] * self.T,
                                                                            self.num_dim + self.N_x)
        # memory free
        state_matrix = None
        gc.collect()
        # get the prediction on the test set
        y_pred = ridge_classifier.predict(new_x_test)
        # reconstruct the test predictions
        y_pred_reshaped = self.reshape_prediction(y_pred, x_test.shape[0], self.T)

        return y_pred_reshaped



import os
from abc import ABC, abstractmethod

import numpy as np
from sklearn.metrics import confusion_matrix

from useckit.util.dataset import Dataset
from useckit.util.dataset_windowsliced import WindowslicedDataset
from useckit.util.plotting import plot_confusion_matrix
from useckit.util.utils import create_classification_report


class VerificationModel(ABC):

    @abstractmethod
    def verify(self, samples: np.ndarray, identity_claims: np.ndarray) -> np.ndarray:
        # identity_claims should be a 1-dim array of ints corresponding to the number transformed by the dataset used
        # in training. It should have the same size as the first dimension of samples
        # the result is a boolean array of same size as identity_claims
        pass


def perform_verification_evaluation(verification_model: VerificationModel, dataset: Dataset, output_folder: str):
    filename_prefix = 'windowslices-' if isinstance(dataset, WindowslicedDataset) else ''
    _verify_data_labels(dataset, verification_model, output_folder, filename_prefix, reverse_windowslicing=False)

    if isinstance(dataset, WindowslicedDataset):
        _verify_data_labels(dataset, verification_model, output_folder, '', reverse_windowslicing=True)


def _verify_data_labels(dataset: Dataset, verification_model: VerificationModel, output_folder: str,
                        filename_prefix: str, reverse_windowslicing: bool):
    testset_data, testset_labels = dataset.testset_matching_data, dataset.testset_matching_labels
    enrollment_labels = set(dataset.testset_enrollment_labels)
    if reverse_windowslicing:
        assert isinstance(dataset, WindowslicedDataset)
        number_verifications = len(dataset.testset_matching_sliceorigin_mask)
    else:
        number_verifications = len(testset_data)

    total_ground_truth = np.zeros(len(enrollment_labels) * number_verifications, dtype=bool)
    total_predictions = np.zeros(len(enrollment_labels) * number_verifications, dtype=bool)
    total_index = 0

    # make all predictions in one call to prediction, then later use the label_claims_array to split it back into the
    # predictions for the individual enrolment labels
    testset_data_array = []
    label_claims_array = []
    for enrolment_label in enrollment_labels:
        label_claims: np.ndarray = np.ones(shape=testset_labels.shape, dtype=int) * enrolment_label
        testset_data_array.append(testset_data)
        label_claims_array.append(label_claims)

    testset_data_array = np.concatenate(testset_data_array)
    label_claims_array = np.concatenate(label_claims_array)
    predictions_array = verification_model.verify(testset_data_array, label_claims_array)

    for enrolment_label in enrollment_labels:
        testset_labels: np.ndarray = testset_labels
        label_claims: np.ndarray = np.ones(shape=testset_labels.shape, dtype=int) * enrolment_label

        if reverse_windowslicing:
            ground_truth_slices: np.ndarray = testset_labels == label_claims
            predictions_slices: np.ndarray = predictions_array[label_claims_array == enrolment_label]
            predictions, ground_truth = dataset.apply_voting_for_testset_matching_slices(predictions_slices,
                                                                                         ground_truth_slices)
        else:
            ground_truth: np.ndarray = testset_labels == label_claims
            predictions: np.ndarray = predictions_array[label_claims_array == enrolment_label]

        total_ground_truth[total_index:total_index + len(ground_truth)] = ground_truth
        total_predictions[total_index:total_index + len(predictions)] = predictions
        total_index += len(ground_truth)

        label_reverse_transformed = str(dataset.reverse_label_transform(np.array([enrolment_label]))[0])
        label_sub_dir = os.path.join(output_folder, label_reverse_transformed)
        os.makedirs(label_sub_dir, exist_ok=True)

        _output_results(predictions, ground_truth, label_sub_dir, f'{filename_prefix}{label_reverse_transformed}-')

    _output_results(total_predictions, total_ground_truth, output_folder, f'{filename_prefix}total-')


def _output_results(predictions: np.ndarray, ground_truth: np.ndarray, output_folder: str,
                    filename_prefix: str = ''):
    os.makedirs(output_folder, exist_ok=True)
    cm = confusion_matrix(ground_truth, predictions, labels=[True, False])
    labels = ['accept', 'reject']
    create_classification_report(cm, labels, output_folder, filename=f'{filename_prefix}classification-report')
    plot_confusion_matrix(cm, target_names=labels, path=output_folder, normalize=False,
                          filename=f'{filename_prefix}confusion-matrix.pdf')



from useckit.models.model_builder import ModelBuilder, ModelArchitecture, Paradigm
from useckit.util.dataset import Dataset
import unittest
import keras
import numpy
from sklearn.metrics import matthews_corrcoef, accuracy_score

class TestModelFactory(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        self.unittest_execution_counter = 1

    def load_prep_data(self):
        (trainset_data, trainset_labels), (testset_data, testset_labels) = keras.datasets.fashion_mnist.load_data()
        trainset_data = (trainset_data / 255).astype(numpy.float16)
        testset_data = (testset_data / 255).astype(numpy.float16)
        class_names = ["T-shirt/top", "Trouser", "Pullover", "Dress", "Coat", "Sandal", "Shirt", "Sneaker", "Bag", "Ankle boot"]
        return Dataset(
            trainset_data=trainset_data,
            trainset_labels=trainset_labels,
            testset_enrollment_data=trainset_data,
            testset_enrollment_labels=trainset_labels,
            testset_matching_data=testset_data,
            testset_matching_labels=testset_labels
        ), class_names

    def test_bv_one_one(self):
        dataset, class_names = self.load_prep_data()
        class_names = class_names[:2]
        train_indices = numpy.where(dataset.trainset_labels >= 2)
        test_indices = numpy.where(dataset.testset_matching_labels >= 2)
        
        dataset.trainset_labels = numpy.delete(dataset.trainset_labels, train_indices)
        dataset.testset_matching_labels = numpy.delete(dataset.testset_matching_labels, test_indices)
        dataset.trainset_data = numpy.delete(dataset.trainset_data, train_indices, axis=0)
        dataset.testset_matching_data = numpy.delete(dataset.testset_matching_data, test_indices, axis=0)
        
        builder = ModelBuilder(dataset)
        model = builder.create_model(ModelArchitecture.MLP, Paradigm.BV)
        
        model.fit(dataset.trainset_data, dataset.trainset_labels, verbose=2)
        
        y_pred = numpy.round(model.predict(dataset.testset_matching_data, verbose=2))
        y_true = dataset.testset_matching_labels
        
        mcc = matthews_corrcoef(y_true, y_pred)
        test_acc = accuracy_score(y_true, y_pred)

        print("Test accuracy:", test_acc, "Matthews Corrcoef:", mcc)
        
        assert mcc >= 0.85, f"Expected MCC to be >= .85, but encountered: {mcc}"

    def test_bv_one_multiple(self):
        dataset, class_names = self.load_prep_data()
        class_names = class_names[:1] + ["Other"]
        dataset.trainset_labels[dataset.trainset_labels > 0] = 1
        dataset.testset_matching_labels[dataset.testset_matching_labels > 0] = 1
        
        builder = ModelBuilder(dataset)
        model = builder.create_model(ModelArchitecture.RESNET, "binary_verification")
        
        model.fit(dataset.trainset_data, dataset.trainset_labels, verbose=2)
        
        y_pred = numpy.round(model.predict(dataset.testset_matching_data, verbose=2))
        y_true = dataset.testset_matching_labels
        
        mcc = matthews_corrcoef(y_true, y_pred)
        test_acc = accuracy_score(y_true, y_pred)

        print("Test accuracy:", test_acc, "Matthews Corrcoef:", mcc)
        
        assert mcc >= 0.65, f"Expected MCC to be >= .65, but encountered: {mcc}"


if __name__ == '__main__':
    unittest.main()



import logging

import tensorflow as tf


class GroupNormalization(tf.keras.layers.Layer):
    """Group normalization layer.

    Source: "Group Normalization" (Yuxin Wu & Kaiming He, 2018)
    https://arxiv.org/abs/1803.08494

    Group Normalization divides the channels into groups and computes
    within each group the mean and variance for normalization.
    Empirically, its accuracy is more stable than batch norm in a wide
    range of small batch sizes, if learning rate is adjusted linearly
    with batch sizes.

    Relation to Layer Normalization:
    If the number of groups is set to 1, then this operation becomes identical
    to Layer Normalization.

    Relation to Instance Normalization:
    If the number of groups is set to the
    input dimension (number of groups is equal
    to number of channels), then this operation becomes
    identical to Instance Normalization.

    Args:
        groups: Integer, the number of groups for Group Normalization.
            Can be in the range [1, N] where N is the input dimension.
            The input dimension must be divisible by the number of groups.
            Defaults to 32.
        axis: Integer, the axis that should be normalized.
        epsilon: Small float added to variance to avoid dividing by zero.
        center: If True, add offset of `beta` to normalized tensor.
            If False, `beta` is ignored.
        scale: If True, multiply by `gamma`.
            If False, `gamma` is not used.
        beta_initializer: Initializer for the beta weight.
        gamma_initializer: Initializer for the gamma weight.
        beta_regularizer: Optional regularizer for the beta weight.
        gamma_regularizer: Optional regularizer for the gamma weight.
        beta_constraint: Optional constraint for the beta weight.
        gamma_constraint: Optional constraint for the gamma weight.

    Input shape:
        Arbitrary. Use the keyword argument `input_shape`
        (tuple of integers, does not include the samples axis)
        when using this layer as the first layer in a model.

    Output shape:
        Same shape as input.
    """

    def __init__(
        self,
        groups: int = 32,
        axis: int = -1,
        epsilon: float = 1e-3,
        center: bool = True,
        scale: bool = True,
        beta_initializer="zeros",
        gamma_initializer="ones",
        beta_regularizer=None,
        gamma_regularizer=None,
        beta_constraint=None,
        gamma_constraint=None,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.supports_masking = True
        self.groups = groups
        self.axis = axis
        self.epsilon = epsilon
        self.center = center
        self.scale = scale
        self.beta_initializer = tf.keras.initializers.get(beta_initializer)
        self.gamma_initializer = tf.keras.initializers.get(gamma_initializer)
        self.beta_regularizer = tf.keras.regularizers.get(beta_regularizer)
        self.gamma_regularizer = tf.keras.regularizers.get(gamma_regularizer)
        self.beta_constraint = tf.keras.constraints.get(beta_constraint)
        self.gamma_constraint = tf.keras.constraints.get(gamma_constraint)
        self._check_axis()

    def build(self, input_shape):

        self._check_if_input_shape_is_none(input_shape)
        self._set_number_of_groups_for_instance_norm(input_shape)
        self._check_size_of_dimensions(input_shape)
        self._create_input_spec(input_shape)

        self._add_gamma_weight(input_shape)
        self._add_beta_weight(input_shape)
        self.built = True
        super().build(input_shape)

    def call(self, inputs):

        input_shape = tf.keras.backend.int_shape(inputs)
        tensor_input_shape = tf.shape(inputs)

        reshaped_inputs, group_shape = self._reshape_into_groups(
            inputs, input_shape, tensor_input_shape
        )

        normalized_inputs = self._apply_normalization(reshaped_inputs, input_shape)

        is_instance_norm = (input_shape[self.axis] // self.groups) == 1
        if not is_instance_norm:
            outputs = tf.reshape(normalized_inputs, tensor_input_shape)
        else:
            outputs = normalized_inputs

        return outputs

    def get_config(self):
        config = {
            "groups": self.groups,
            "axis": self.axis,
            "epsilon": self.epsilon,
            "center": self.center,
            "scale": self.scale,
            "beta_initializer": tf.keras.initializers.serialize(self.beta_initializer),
            "gamma_initializer": tf.keras.initializers.serialize(
                self.gamma_initializer
            ),
            "beta_regularizer": tf.keras.regularizers.serialize(self.beta_regularizer),
            "gamma_regularizer": tf.keras.regularizers.serialize(
                self.gamma_regularizer
            ),
            "beta_constraint": tf.keras.constraints.serialize(self.beta_constraint),
            "gamma_constraint": tf.keras.constraints.serialize(self.gamma_constraint),
        }
        base_config = super().get_config()
        return {**base_config, **config}

    def _reshape_into_groups(self, inputs, input_shape, tensor_input_shape):

        group_shape = [tensor_input_shape[i] for i in range(len(input_shape))]
        is_instance_norm = (input_shape[self.axis] // self.groups) == 1
        if not is_instance_norm:
            group_shape[self.axis] = input_shape[self.axis] // self.groups
            group_shape.insert(self.axis, self.groups)
            group_shape = tf.stack(group_shape)
            reshaped_inputs = tf.reshape(inputs, group_shape)
            return reshaped_inputs, group_shape
        else:
            return inputs, group_shape

    def _apply_normalization(self, reshaped_inputs, input_shape):

        group_shape = tf.keras.backend.int_shape(reshaped_inputs)
        group_reduction_axes = list(range(1, len(group_shape)))
        is_instance_norm = (input_shape[self.axis] // self.groups) == 1
        if not is_instance_norm:
            axis = -2 if self.axis == -1 else self.axis - 1
        else:
            axis = -1 if self.axis == -1 else self.axis - 1
        group_reduction_axes.pop(axis)

        mean, variance = tf.nn.moments(
            reshaped_inputs, group_reduction_axes, keepdims=True
        )

        gamma, beta = self._get_reshaped_weights(input_shape)
        normalized_inputs = tf.nn.batch_normalization(
            reshaped_inputs,
            mean=mean,
            variance=variance,
            scale=gamma,
            offset=beta,
            variance_epsilon=self.epsilon,
        )
        return normalized_inputs

    def _get_reshaped_weights(self, input_shape):
        broadcast_shape = self._create_broadcast_shape(input_shape)
        gamma = None
        beta = None
        if self.scale:
            gamma = tf.reshape(self.gamma, broadcast_shape)

        if self.center:
            beta = tf.reshape(self.beta, broadcast_shape)
        return gamma, beta

    def _check_if_input_shape_is_none(self, input_shape):
        dim = input_shape[self.axis]
        if dim is None:
            raise ValueError(
                "Axis " + str(self.axis) + " of "
                                           "input tensor should have a defined dimension "
                                           "but the layer received an input with shape " + str(input_shape) + "."
            )

    def _set_number_of_groups_for_instance_norm(self, input_shape):
        dim = input_shape[self.axis]

        if self.groups == -1:
            self.groups = dim

    def _check_size_of_dimensions(self, input_shape):

        dim = input_shape[self.axis]
        if dim < self.groups:
            raise ValueError(
                "Number of groups (" + str(self.groups) + ") cannot be "
                                                          "more than the number of channels (" + str(dim) + ")."
            )

        if dim % self.groups != 0:
            raise ValueError(
                "Number of groups (" + str(self.groups) + ") must be a "
                                                          "multiple of the number of channels (" + str(dim) + ")."
            )

    def _check_axis(self):

        if self.axis == 0:
            raise ValueError(
                "You are trying to normalize your batch axis. Do you want to "
                "use tf.layer.batch_normalization instead"
            )

    def _create_input_spec(self, input_shape):

        dim = input_shape[self.axis]
        self.input_spec = tf.keras.layers.InputSpec(
            ndim=len(input_shape), axes={self.axis: dim}
        )

    def _add_gamma_weight(self, input_shape):

        dim = input_shape[self.axis]
        shape = (dim,)

        if self.scale:
            self.gamma = self.add_weight(
                shape=shape,
                name="gamma",
                initializer=self.gamma_initializer,
                regularizer=self.gamma_regularizer,
                constraint=self.gamma_constraint,
            )
        else:
            self.gamma = None

    def _add_beta_weight(self, input_shape):

        dim = input_shape[self.axis]
        shape = (dim,)

        if self.center:
            self.beta = self.add_weight(
                shape=shape,
                name="beta",
                initializer=self.beta_initializer,
                regularizer=self.beta_regularizer,
                constraint=self.beta_constraint,
            )
        else:
            self.beta = None

    def _create_broadcast_shape(self, input_shape):
        broadcast_shape = [1] * len(input_shape)
        is_instance_norm = (input_shape[self.axis] // self.groups) == 1
        if not is_instance_norm:
            broadcast_shape[self.axis] = input_shape[self.axis] // self.groups
            broadcast_shape.insert(self.axis, self.groups)
        else:
            broadcast_shape[self.axis] = self.groups
        return broadcast_shape


class InstanceNormalization(GroupNormalization):
    """Instance normalization layer.

    Instance Normalization is an specific case of ```GroupNormalization```since
    it normalizes all features of one channel. The Groupsize is equal to the
    channel size. Empirically, its accuracy is more stable than batch norm in a
    wide range of small batch sizes, if learning rate is adjusted linearly
    with batch sizes.

    Arguments
        axis: Integer, the axis that should be normalized.
        epsilon: Small float added to variance to avoid dividing by zero.
        center: If True, add offset of `beta` to normalized tensor.
            If False, `beta` is ignored.
        scale: If True, multiply by `gamma`.
            If False, `gamma` is not used.
        beta_initializer: Initializer for the beta weight.
        gamma_initializer: Initializer for the gamma weight.
        beta_regularizer: Optional regularizer for the beta weight.
        gamma_regularizer: Optional regularizer for the gamma weight.
        beta_constraint: Optional constraint for the beta weight.
        gamma_constraint: Optional constraint for the gamma weight.

    Input shape
        Arbitrary. Use the keyword argument `input_shape`
        (tuple of integers, does not include the samples axis)
        when using this layer as the first layer in a model.

    Output shape
        Same shape as input.

    References
        - [Instance Normalization: The Missing Ingredient for Fast Stylization]
        (https://arxiv.org/abs/1607.08022)
    """

    def __init__(self, **kwargs):
        if "groups" in kwargs:
            logging.warning("The given value for groups will be overwritten.")

        kwargs["groups"] = -1
        super().__init__(**kwargs)



import os
import pickle

import numpy as np
from sklearn.ensemble import RandomForestClassifier

from .tsc_prediction_model_base import TSCBasePredictionModel
from ....util.dataset import Dataset


class ClassificationScikitPredictionModel(TSCBasePredictionModel):
    def __init__(self,
                 scikit_classifier=RandomForestClassifier(),
                 sample_weights=None,
                 verbose: bool = True,
                 output_dir: str = "scikit_classifier_out"):
        super().__init__(output_dir=output_dir, verbose=verbose)
        self.classifier = scikit_classifier
        self.sample_weights = sample_weights
        self.model_description = str(type(self.classifier))

    def persist_model_trainables(self, path_like: str = 'saved_trainable_values', **kwargs):
        with open(os.path.join(self.output_dir, path_like), "wb") as file:
            pickle.dump(self.classifier, file)

    def restore_model_trainables(self, path_like, dataset: Dataset = None, **kwargs):
        with open(path_like, "rb") as file:
            self.classifier = pickle.load(file)

    def fit(self, dataset: Dataset, *args, **kwargs):
        x_train, y_train = dataset.trainset_data, dataset.trainset_labels
        x_train = x_train.reshape((x_train.shape[0], np.prod(x_train.shape[1:])))
        self.classifier = self.classifier.fit(x_train, y_train, sample_weight=self.sample_weights)

    def predict(self, x_test, y_true, x_train, y_train, y_test, return_df_metrics=True):
        x_test = x_test.reshape((x_test.shape[0], np.prod(x_test.shape[1:])))
        return self.classifier.predict(x_test)



import json
import os.path
import sys
from collections import OrderedDict

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from scipy.stats import hmean
from sklearn.metrics import accuracy_score
from sklearn.metrics import precision_score
from sklearn.metrics import recall_score


def _unison_shuffle_np_arrays_3(a: np.ndarray, b: np.ndarray, c: np.ndarray):
    assert len(a) == len(b) == len(c)
    p = np.random.permutation(len(a))
    return a[p], b[p], c[p]


def create_window_slices(data: np.ndarray,
                         labels: np.ndarray,
                         stride: int,
                         slice_size: int,
                         shuffle: bool = True) -> [np.ndarray, np.ndarray]:
    if not data.shape[0] == labels.shape[0]:
        raise TypeError("Error in AutoEvaluator::_create_slices(). Data and Labels differ in size.")

    if not len(data.shape) <= 3:
        raise TypeError("Error in AutoEvaluator::_create_slices(). Dimension not supported.")

    ret_sliced_data, ret_sliced_labels, ret_sample_origin = [], [], []

    for sample_id, (array, label) in enumerate(zip(data, labels)):
        array_slices, label_slices, sample_origin = [], [], []

        lost_samples_cnt, added_cnt = 0, 0

        for step_idx in range(0, array.shape[0], stride):
            arr = array[step_idx:step_idx + slice_size]

            if not arr.any():
                pass  # do not append if slice consists only of zero
            else:
                array_slices.append(arr)
                label_slices.append(label)
                sample_origin.append(sample_id)

        for i in range(len(array_slices)):
            if array_slices[i].shape == array_slices[0].shape:
                ret_sliced_data.append(array_slices[i])
                ret_sliced_labels.append(label_slices[i])
                ret_sample_origin.append(sample_origin[i])
                added_cnt += 1
            else:
                lost_samples_cnt += 1

        if not all(s.shape == array_slices[0].shape for s in array_slices):
            print('Warning: `slices` array created in AutoEvaluator::_create_slices() is ragged. '
                  'Check parameters stride, slice_size and shape of array. Lost',
                  lost_samples_cnt, 'samples and created', added_cnt,
                  'new samples. Check your slicing window settings.', file=sys.stderr)

    ret_sliced_data_np = np.array(ret_sliced_data)
    ret_sliced_labels_np = np.array(ret_sliced_labels)
    sample_origin_np = np.array(ret_sample_origin)

    print('Info: Slicing window data consumes', round(ret_sliced_data_np.nbytes / 1024 / 1024, 2), 'mb.',
          file=sys.stderr)

    # TODO check that supressing null arrays does not result in a skewed data set

    if shuffle:
        return _unison_shuffle_np_arrays_3(ret_sliced_data_np, ret_sliced_labels_np, sample_origin_np)
    else:
        return ret_sliced_data_np, ret_sliced_labels_np, sample_origin_np


def make_some_intelligent_noise(labels: int = 4, shape: tuple = (8, 100, 100, 4), noisiness: float = 0.5):
    biases = np.array([i / labels for i in range(labels)]) + (1 / (2 * labels))
    x = []
    y = []
    shape = list(shape)
    shape[0] = int(shape[0] / labels)

    for i, b in enumerate(biases):
        x.extend(np.random.normal(b, noisiness, shape))
        y.extend(np.ones((shape[0],)) * i)

    x = np.array(x)
    y = [f'User {int(y_elem)}' for y_elem in y]
    y = np.array(y)
    x = np.clip(x, -1, 1)
    return x, y


def contrastive_make_pairs(x, y):
    """Creates a tuple containing image pairs with corresponding label.

    Arguments:
        x: List containing images, each index in this list corresponds to one image.
        y: List containing labels, each label with datatype of `int`.

    Returns:
        Tuple containing two numpy arrays as (pairs_of_samples, labels),
        where pairs_of_samples' shape is (2len(x), 2,n_features_dims) and
        labels are a binary array of shape (2len(x)).
    """
    x = np.array(x)
    y = np.array(y)
    assert x.shape[0] == y.shape[0]
    y_set = set(y)

    pairs = []
    labels = []

    matching_masks = {}
    non_matching_masks = {}
    for label in y_set:
        match = (y == label)
        matching_masks[label] = match
        non_matching_masks[label] = np.invert(match)

    for i in range(len(x)):
        x1 = x[i]
        label = y[i]

        # generate matching pair
        matches = x[matching_masks[label]]
        i = np.random.randint(0, len(matches))
        x2 = matches[i]
        pairs += [[x1, x2]]
        labels += [0]

        # generate non-matching pair
        matches = x[non_matching_masks[label]]
        i = np.random.randint(0, len(matches))
        x2 = matches[i]
        pairs += [[x1, x2]]
        labels += [1]

    return np.array(pairs), np.array(labels).astype("float32")


def triplet_make_pairs_random(x, y):
    """Creates a tuple containing sample pairs with corresponding label. Both positive and negative samples are selected randomly.

    Arguments:
        x: Numpy array containing samples, each index in this list corresponds to one sample.
        y: Numpy array containing samples, each label with datatype of `int`.

    Returns:
        Tuple containing two numpy arrays as (pairs_of_samples, labels),
        where pairs_of_samples' shape is len(x) containing the base (0), positive (1) and negative sample (2) and
        labels are an array of shape len(x) containing the label of each base sample.
    """

    x = np.array(x)
    y = np.array(y)
    assert x.shape[0] == y.shape[0]
    y_set = set(y)

    pairs = []  # output: sample pairs for the siamese net (anchor, positive, negative)
    labels = []  # output: labels of the anchor sample per pair

    # indexes of the sample positions per label
    matching_masks = {}
    non_matching_masks = {}
    for label in y_set:
        matching_masks[label] = np.where(y == label)  # returns the indices of the current label
        non_matching_masks[label] = np.where(y != label)  # returns the indices of all other labels

    # iterate over all provided samples to build one pair per sample
    for i in range(x.shape[0]):
        # get anchor sample and its label
        anchor = x[i]
        anchor_label = y[i]

        # generate the positive pair -> another sample having the same label as the base sample
        positive_index = np.random.choice(matching_masks[anchor_label][0])
        while positive_index == i:  # the positive sample may not be the same as the base sample
            positive_index = np.random.choice(matching_masks[anchor_label][0])
        positive = x[positive_index]

        # generate non-matching pair
        negative = x[np.random.choice(non_matching_masks[anchor_label][0])]

        # append the sample pair to the output arrays
        pairs += [[anchor, positive, negative]]
        labels += [anchor_label]

    return np.array(pairs), np.array(labels).astype("float32")


def triplet_make_pairs_exhaust_positives(x, y):
    """Creates a tuple containing sample pairs with corresponding label. The positive values are exhaustedly matched,
     while negative samples are selected randomly.

    Arguments:
        x: Numpy array containing samples, each index in this list corresponds to one sample.
        y: Numpy array containing samples, each label with datatype of `int`.

    Returns:
        Tuple containing two numpy arrays as (pairs_of_samples, labels),
        where pairs_of_samples' shape is (len(samples per label) * (len(samples per label)-1) containing the anchor (0), positive (1) and negative sample (2) and
        labels are an array of shape (len(samples per label) * (len(samples per label)-1) containing the label of each base sample.
    """

    x = np.array(x)
    y = np.array(y)
    assert x.shape[0] == y.shape[0]
    y_set = set(y)

    pairs = []  # output: sample pairs for the siamese net (anchor, positive, negative)
    labels = []  # output: labels of the anchor sample per pair

    matching_masks = {}
    non_matching_masks = {}
    for label in y_set:
        matching_masks[label] = np.where(y == label)  # returns the indices of the current label
        non_matching_masks[label] = np.where(y != label)  # returns the indices of all other labels

    for i in range(x.shape[0]):
        anchor = x[i]
        anchor_label = y[i]

        for j in range(len(matching_masks[anchor_label][0])):

            if matching_masks[anchor_label][0][
                j] != i:  # ensure that the base and positive sample are not equal, skip if true
                positive = x[matching_masks[anchor_label][0][j]]

                # generate non-matching pair by choosing a random sample from the samples not being the anchor
                negative = x[np.random.choice(non_matching_masks[anchor_label][0])]

                pairs += [[anchor, positive, negative]]
                labels += [anchor_label]

    return np.array(pairs), np.array(labels).astype("float32")


def triplet_make_pairs_exhaust_all(x, y):
    """Creates a tuple containing sample pairs with corresponding label. The positive values are exhaustedly matched,
         for the negative samples the matches are choosen equally between the other labels and are each only used once
          per anchor.

        Arguments:
            x: Numpy array containing samples, each index in this list corresponds to one sample.
            y: Numpy array containing samples, each label with datatype of `int`.

        Returns:
            Tuple containing two numpy arrays as (pairs_of_samples, labels),
            where pairs_of_samples' shape is (len(samples per label)^2 * (len(samples per label)-1) containing the anchor (0), positive (1) and negative sample (2) and
            labels are an array of shape (len(samples per label)^2 * (len(samples per label)-1) containing the label of each anchor sample.
        """

    x = np.array(x)
    y = np.array(y)
    assert x.shape[0] == y.shape[0]
    y_set = set(y)

    pairs = []  # output: sample pairs for the siamese net (anchor, positive, negative)
    labels = []  # output: labels of the anchor sample per pair

    matching_masks = {}
    non_matching_indexes = {}  # contains indexes in this implementation
    for label in y_set:
        matching_masks[label] = np.where(y == label)  # returns the indices of the current label

    for i in range(x.shape[0]):
        anchor = x[i]
        anchor_label = y[i]

        non_matching_masks = []  # returns the indices of all other labels in separate arrays
        for other_label in y_set:
            if other_label != anchor_label:
                non_matching_masks.append(np.where(y == other_label))

        for j in range(len(matching_masks[label][0])):

            if matching_masks[anchor_label][0][
                j] != i:  # ensure that the base and positive sample are not equal, skip if true
                # generate positive pair
                positive = x[matching_masks[anchor_label][0][j]]

                # generate non-matching pair
                for k in range(len(non_matching_masks)):
                    rand_index = np.random.randint(0, len(non_matching_masks[k][0]))
                    negative = x[non_matching_masks[k][0][rand_index]]

                    # delete taken sample from mask to ensure that it is not taken a second time
                    non_matching_masks[k] = tuple([np.delete(non_matching_masks[k][0], rand_index)])

                    pairs += [[anchor, positive, negative]]
                    labels += [anchor_label]

    return np.array(pairs), np.array(labels).astype("float32")


def calculate_metrics(y_true, y_pred, duration, y_true_val=None, y_pred_val=None):
    res = pd.DataFrame(data=np.zeros((1, 4), dtype=np.float64), index=[0],
                       columns=['precision', 'accuracy', 'recall', 'duration'])
    res['precision'] = precision_score(y_true, y_pred, average='macro')
    res['accuracy'] = accuracy_score(y_true, y_pred)

    if not y_true_val is None:
        # this is useful when transfer learning is used with cross validation
        res['accuracy_val'] = accuracy_score(y_true_val, y_pred_val)

    res['recall'] = recall_score(y_true, y_pred, average='macro')
    res['duration'] = duration
    return res


def plot_epochs_metric(hist, file_name, metric='loss'):
    plt.figure()
    plt.plot(hist.history[metric])
    plt.plot(hist.history['val_' + metric])
    plt.title('model ' + metric)
    plt.ylabel(metric, fontsize='large')
    plt.xlabel('epoch', fontsize='large')
    plt.legend(['train', 'val'], loc='upper left')
    plt.savefig(file_name, bbox_inches='tight')
    plt.close()


def save_logs(output_directory, hist, y_pred, y_true, duration, lr=True, y_true_val=None, y_pred_val=None):
    hist_df = pd.DataFrame(hist.history)
    hist_df.to_csv(os.path.join(output_directory, 'history.csv'), index=False)

    df_metrics = calculate_metrics(y_true, y_pred, duration, y_true_val, y_pred_val)
    df_metrics.to_csv(os.path.join(output_directory, 'df_metrics.csv'), index=False)

    index_best_model = hist_df['loss'].idxmin()
    row_best_model = hist_df.loc[index_best_model]

    df_best_model = pd.DataFrame(data=np.zeros((1, 6), dtype=np.float64), index=[0],
                                 columns=['best_model_train_loss', 'best_model_val_loss', 'best_model_train_acc',
                                          'best_model_val_acc', 'best_model_learning_rate', 'best_model_nb_epoch'])

    df_best_model['best_model_train_loss'] = row_best_model['loss']
    df_best_model['best_model_val_loss'] = row_best_model['val_loss']
    df_best_model['best_model_train_acc'] = row_best_model['accuracy']
    df_best_model['best_model_val_acc'] = row_best_model['val_accuracy']
    if lr == True:
        df_best_model['best_model_learning_rate'] = row_best_model['lr']
    df_best_model['best_model_nb_epoch'] = index_best_model

    df_best_model.to_csv(os.path.join(output_directory, 'df_best_model.csv'), index=False)

    # for FCN there is no hyperparameters fine tuning - everything is static in code

    # plot losses
    plot_epochs_metric(hist, os.path.join(output_directory, 'epochs_loss.png'))

    return df_metrics


def save_test_duration(file_name, test_duration):
    res = pd.DataFrame(data=np.zeros((1, 1), dtype=np.float64), index=[0],
                       columns=['test_duration'])
    res['test_duration'] = test_duration
    res.to_csv(file_name, index=False)


def create_directory(directory_path):
    if os.path.exists(directory_path):
        return None
    else:
        try:
            os.makedirs(directory_path)
        except:
            # in case another machine created the path meanwhile !:(
            return None
        return directory_path


def save_logs_t_leNet(output_directory, hist, y_pred, y_true, duration):
    hist_df = pd.DataFrame(hist.history)
    hist_df.to_csv(os.path.join(output_directory, 'history.csv'), index=False)

    df_metrics = calculate_metrics(y_true, y_pred, duration)
    df_metrics.to_csv(os.path.join(output_directory, 'df_metrics.csv'), index=False)

    index_best_model = hist_df['loss'].idxmin()
    row_best_model = hist_df.loc[index_best_model]

    df_best_model = pd.DataFrame(data=np.zeros((1, 6), dtype=np.float64), index=[0],
                                 columns=['best_model_train_loss', 'best_model_val_loss', 'best_model_train_acc',
                                          'best_model_val_acc', 'best_model_learning_rate', 'best_model_nb_epoch'])

    df_best_model['best_model_train_loss'] = row_best_model['loss']
    df_best_model['best_model_val_loss'] = row_best_model['val_loss']
    df_best_model['best_model_train_acc'] = row_best_model['accuracy']
    df_best_model['best_model_val_acc'] = row_best_model['val_accuracy']
    df_best_model['best_model_nb_epoch'] = index_best_model

    df_best_model.to_csv(os.path.join(output_directory, 'df_best_model.csv'), index=False)

    # plot losses
    plot_epochs_metric(hist, os.path.join(output_directory, 'epochs_loss.png'))


def legacy_serialize_classification_report(arr1, arr2, output_folder, filename='classification-report.txt'):
    """Writes a classification_report from sklearn to file as .txt and .json."""
    from sklearn.metrics import classification_report
    with open(os.path.join(output_folder, filename), 'w') as f:
        f.write('Classification Report\n')
        f.write(classification_report(arr1, arr2))
    with open(os.path.join(output_folder, filename.replace('.txt', '.json')), 'w') as f2:
        f2.write(json.dumps(classification_report(arr1, arr2, output_dict=True)))


def create_classification_report(cm, labels, output_folder, filename='classification-report'):
    assert len(cm) > 1
    binary = len(cm) == 2
    if binary:
        report, report_dict = _create_binary_report(cm, labels)
    else:
        report, report_dict = _create_multiclass_report(cm, labels)
    with open(os.path.join(output_folder, filename + ".txt"), 'w') as f:
        f.write('Classification Report\n')
        f.write(report)
    with open(os.path.join(output_folder, filename + ".json"), 'w') as f2:
        json.dump(report_dict, f2, indent=4)


def _create_multiclass_report(cm, labels):
    trans_cm = np.transpose(cm)
    total_predictions = np.sum(cm)
    assert total_predictions > 0
    assert len(labels) == len(cm) == len(trans_cm)

    max_len_labels = 0
    for label in labels:
        if len(label) > max_len_labels:
            max_len_labels = len(label)
    if len("macro avg") > max_len_labels:
        max_len_labels = len("macro avg")

    headers_list = ["precision", "recall", "f1-score", "prevalence", "bias"]
    header_row = "{:>{mll}s} " + " {:^10s}" * len(headers_list)
    header_row = header_row.format("", *headers_list, mll=max_len_labels) + "\n\n"
    report = header_row

    true_predictions = 0
    metric_averages_dict = dict()
    for header in headers_list:
        metric_averages_dict[header] = 0
    label_metric_dicts = dict()
    for i, label in enumerate(labels):
        row = "{:>{mll}} " + " {:^10.2f}" * len(headers_list) + "\n"
        support = np.sum(cm[i])
        predicted_support = np.sum(trans_cm[i])
        prevalence = support / total_predictions
        bias = predicted_support / total_predictions
        recall = 0 if support == 0 else cm[i][i] / support
        precision = 0 if predicted_support == 0 else trans_cm[i][i] / predicted_support
        f_1 = hmean([precision, recall])
        row = row.format(label, precision, recall, f_1, prevalence, bias, mll=max_len_labels)
        report += row

        l_dict = dict()
        l_dict["precision"] = precision
        metric_averages_dict["precision"] += precision
        l_dict["recall"] = recall
        metric_averages_dict["recall"] += recall
        l_dict["f1-score"] = f_1
        metric_averages_dict["f1-score"] += f_1
        l_dict["prevalence"] = prevalence
        metric_averages_dict["prevalence"] += prevalence
        l_dict["bias"] = bias
        metric_averages_dict["bias"] += bias
        label_metric_dicts[label] = l_dict

        true_predictions += cm[i][i]

    report += "\n"

    metric_averages_dict["precision"] /= len(labels)
    metric_averages_dict["recall"] /= len(labels)
    metric_averages_dict["f1-score"] /= len(labels)
    metric_averages_dict["prevalence"] /= len(labels)
    metric_averages_dict["bias"] /= len(labels)

    row = "{:>{mll}} " + " {:^10.2f}" * len(headers_list) + "\n\n"
    row = row.format("macro avg",
                     metric_averages_dict["precision"],
                     metric_averages_dict["recall"],
                     metric_averages_dict["f1-score"],
                     metric_averages_dict["prevalence"],
                     metric_averages_dict["bias"],
                     mll=max_len_labels)
    report += row

    row_format = "{:>{mll}}  {:^10.2f}" + "\n"
    accuracy = true_predictions / total_predictions
    row = row_format.format("accuracy", accuracy, mll=max_len_labels)
    report += row

    result_dict = dict()
    result_dict["binary"] = False
    result_dict["metrics_per_label"] = label_metric_dicts
    result_dict["macro_averages"] = metric_averages_dict
    result_dict["accuracy"] = accuracy
    result_dict["labels"] = labels.tolist() if type(labels) == np.ndarray else labels
    result_dict["contingency_tabel"] = cm.tolist() if type(cm) == np.ndarray else cm

    return report, result_dict


def _create_binary_report(cm, labels):
    trans_cm = np.transpose(cm)
    total_predictions = np.sum(cm)
    true_predictions = cm[0][0] + cm[1][1]
    assert total_predictions > 0
    assert len(labels) == len(cm) == len(trans_cm) == 2

    headers_list = []
    max_len_labels = 0
    for i, label in enumerate(labels):
        headers_list.append("pred. " + label)
        if len(headers_list[i]) > max_len_labels:
            max_len_labels = len(headers_list[i])

    header_row = "{:>{mll}s} " + " {:>{mll}s}" + " {:<{mll}s}"
    header_row = header_row.format("", *headers_list, mll=max_len_labels) + "\n"
    report = header_row

    for i, label in enumerate(labels):
        row = "{:>{mll}s} " + " {:^{mll}}" + " {:^{mll}}" + "\n"
        row = row.format("true " + label, cm[i][0], cm[i][1], mll=max_len_labels)
        report += row

    report += "\n\n"

    support = np.sum(cm[0])
    predicted_support = np.sum(trans_cm[0])
    prevalence = support / total_predictions
    bias = predicted_support / total_predictions
    recall = 0 if support == 0 else cm[0][0] / support
    precision = 0 if predicted_support == 0 else trans_cm[0][0] / predicted_support
    specificity = 0 if total_predictions - support == 0 else cm[1][1] / (total_predictions - support)
    negative_pred_value = 0 if total_predictions - predicted_support == 0 else \
        cm[1][1] / (total_predictions - predicted_support)
    f_1 = hmean([precision, recall])
    accuracy = true_predictions / total_predictions
    balanced_accuracy = (recall + specificity) / 2
    bookmaker_informedness = recall + specificity - 1
    mcc = 0 if bias == 0 or bias == 1 else \
        bookmaker_informedness * np.sqrt((prevalence - prevalence * prevalence) / (bias - bias * bias))

    result_dict = OrderedDict()
    result_dict["accuracy"] = accuracy
    result_dict["prevalence"] = prevalence
    result_dict["bias"] = bias
    result_dict["recall"] = recall
    result_dict["precision"] = precision
    result_dict["specificity"] = specificity
    result_dict["negative predictive value"] = negative_pred_value
    result_dict["F1-score"] = f_1
    result_dict["balanced accuracy"] = balanced_accuracy
    result_dict["bookmaker informedness"] = bookmaker_informedness
    result_dict["MCC"] = mcc

    for key, value in result_dict.items():
        row = "{:>{ml}s}: " + "{:<6.2f}" + "\n"
        row = row.format(key, value, ml=len("negative predictive value"))
        report += row

    result_dict["binary"] = True
    result_dict["labels"] = labels.tolist() if type(labels) == np.ndarray else labels
    result_dict["contingency_tabel"] = cm.tolist() if type(cm) == np.ndarray else cm

    return report, result_dict


def analyze_samplelength(DATASET_LIST: list):
    """Creates a boxplot and descriptive statistics of a list of samples."""
    from statistics import mean, stdev, median
    samplelen_min, samplelen_max, samplelen_mean, samplelen_sd, samplelen_median = \
        (min(len(x['df']) for x in DATASET_LIST), max(len(x['df']) for x in DATASET_LIST),
         round(mean(len(x['df']) for x in DATASET_LIST)),
         round(stdev(len(x['df']) for x in DATASET_LIST)),
         round(median(len(x['df']) for x in DATASET_LIST)))

    def get_q75_q25(dist):
        return np.percentile(dist, 75), np.percentile(dist, 25)

    samplelen_q75, samplelen_q25 = get_q75_q25([len(x['df']) for x in DATASET_LIST])

    plt.boxplot([[len(x['df']) for x in DATASET_LIST if x['sess'] == 1],
                 [len(x['df']) for x in DATASET_LIST if x['sess'] == 2]], vert=False)
    plt.suptitle('Boxplot of sample lengths for scene "ButtonScene1H".')
    plt.title(f"Global min: {samplelen_min}, "
              f"max: {samplelen_max}, "
              f"M: {samplelen_mean}, "
              f"SD: {samplelen_sd}, "
              f"Med: {samplelen_median}, "
              f"Q25: {samplelen_q25}, "
              f"Q75: {samplelen_q75}.")
    plt.xlabel('Sample length (VR Sample Rate: 72 hz)')
    plt.ylabel('Session')
    plt.grid()
    plt.show()
    return {'min': samplelen_min, 'max': samplelen_max, 'mean': samplelen_mean, 'sd': samplelen_sd,
            'median': samplelen_median, 'q75': samplelen_q75, 'q25': samplelen_q25}



from useckit.paradigms.time_series_classification.prediction_models.instance_normalization_layer import \
    InstanceNormalization
from useckit.paradigms.distance_learning.prediction_models.online_triplet_loss.triplet_semihard_loss_function import \
    TripletSemiHardLoss

import keras
import tensorflow as tf
import numpy as np

from enum import Enum

class ModelArchitecture(Enum):
    CNN_VALID = 0
    CNN_SAME = 1
    FCN = 2
    MLP = 3
    RESNET = 4
    ENCODER = 5
    MCNN = 6
    MCDCNN = 7
    TLENET = 8
    INCEPTION = 9
    TWIESN = 10
    AD = ANOMALY_DETECTION = 11

    @classmethod
    def get_by_string(cls, name: str):
        name = name.upper()
        return cls.__members__.get(name)

class Paradigm(Enum):
    TSC = TIME_SERIES_CLASSIFICATION = 0
    BV = BINARY_VERIFICATION = 1
    CL = CONTRASIVE_LOSS = 2
    TL = TRIPLET_LOSS = 3
    AD = ANOMALY_DETECTION = 4

    @classmethod
    def get_by_string(cls, name: str):
        name = name.upper()
        return cls.__members__.get(name)


def _mb_dl4tsc_cnn_mid(input_layer: keras.layers.Input, padding: str): 
    conv1 = keras.layers.Conv1D(filters=6, kernel_size=7, padding=padding, activation='sigmoid')(input_layer)
    conv1 = keras.layers.AveragePooling1D(pool_size=3)(conv1)

    conv2 = keras.layers.Conv1D(filters=12, kernel_size=7, padding=padding, activation='sigmoid')(conv1)
    conv2 = keras.layers.AveragePooling1D(pool_size=3)(conv2)

    flatten_layer = keras.layers.Flatten()(conv2)
    return flatten_layer


# use with batch_size=16 and epochs=2000
def mb_dl4tsc_fcn_mid(input_layer: keras.layers.Input): 
    conv1 = keras.layers.Conv1D(filters=128, kernel_size=8, padding='same')(input_layer)
    conv1 = keras.layers.BatchNormalization()(conv1)
    conv1 = keras.layers.Activation(activation='relu')(conv1)

    conv2 = keras.layers.Conv1D(filters=256, kernel_size=5, padding='same')(conv1)
    conv2 = keras.layers.BatchNormalization()(conv2)
    conv2 = keras.layers.Activation('relu')(conv2)

    conv3 = keras.layers.Conv1D(128, kernel_size=3, padding='same')(conv2)
    conv3 = keras.layers.BatchNormalization()(conv3)
    conv3 = keras.layers.Activation('relu')(conv3)

    gap_layer = keras.layers.GlobalAveragePooling1D()(conv3)
    return gap_layer


# use with batch_size=16 and epochs=5000
def mb_dl4tsc_mlp_mid(input_layer: keras.layers.Input): 
    # flatten/reshape because when multivariate all should be on the same axis
    input_layer_flattened = keras.layers.Flatten()(input_layer)

    layer_1 = keras.layers.Dropout(0.1)(input_layer_flattened)
    layer_1 = keras.layers.Dense(500, activation='relu')(layer_1)

    layer_2 = keras.layers.Dropout(0.2)(layer_1)
    layer_2 = keras.layers.Dense(500, activation='relu')(layer_2)

    layer_3 = keras.layers.Dropout(0.2)(layer_2)
    layer_3 = keras.layers.Dense(500, activation='relu')(layer_3)
    output_layer = keras.layers.Dropout(0.3)(layer_3)
    return output_layer


# use with batch_size=min(64, train_data/10) and epochs=1500
def mb_dl4tsc_resnet_mid(input_layer: keras.layers.Input): 
    n_feature_maps = 64

    # BLOCK 1
    conv_x = keras.layers.Conv1D(filters=n_feature_maps, kernel_size=8, padding='same')(input_layer)
    conv_x = keras.layers.BatchNormalization()(conv_x)
    conv_x = keras.layers.Activation('relu')(conv_x)

    conv_y = keras.layers.Conv1D(filters=n_feature_maps, kernel_size=5, padding='same')(conv_x)
    conv_y = keras.layers.BatchNormalization()(conv_y)
    conv_y = keras.layers.Activation('relu')(conv_y)

    conv_z = keras.layers.Conv1D(filters=n_feature_maps, kernel_size=3, padding='same')(conv_y)
    conv_z = keras.layers.BatchNormalization()(conv_z)

    # expand channels for the sum
    shortcut_y = keras.layers.Conv1D(filters=n_feature_maps, kernel_size=1, padding='same')(input_layer)
    shortcut_y = keras.layers.BatchNormalization()(shortcut_y)

    output_block_1 = keras.layers.add([shortcut_y, conv_z])
    output_block_1 = keras.layers.Activation('relu')(output_block_1)

    # BLOCK 2

    conv_x = keras.layers.Conv1D(filters=n_feature_maps * 2, kernel_size=8, padding='same')(output_block_1)
    conv_x = keras.layers.BatchNormalization()(conv_x)
    conv_x = keras.layers.Activation('relu')(conv_x)

    conv_y = keras.layers.Conv1D(filters=n_feature_maps * 2, kernel_size=5, padding='same')(conv_x)
    conv_y = keras.layers.BatchNormalization()(conv_y)
    conv_y = keras.layers.Activation('relu')(conv_y)

    conv_z = keras.layers.Conv1D(filters=n_feature_maps * 2, kernel_size=3, padding='same')(conv_y)
    conv_z = keras.layers.BatchNormalization()(conv_z)

    # expand channels for the sum
    shortcut_y = keras.layers.Conv1D(filters=n_feature_maps * 2, kernel_size=1, padding='same')(output_block_1)
    shortcut_y = keras.layers.BatchNormalization()(shortcut_y)

    output_block_2 = keras.layers.add([shortcut_y, conv_z])
    output_block_2 = keras.layers.Activation('relu')(output_block_2)

    # BLOCK 3

    conv_x = keras.layers.Conv1D(filters=n_feature_maps * 2, kernel_size=8, padding='same')(output_block_2)
    conv_x = keras.layers.BatchNormalization()(conv_x)
    conv_x = keras.layers.Activation('relu')(conv_x)

    conv_y = keras.layers.Conv1D(filters=n_feature_maps * 2, kernel_size=5, padding='same')(conv_x)
    conv_y = keras.layers.BatchNormalization()(conv_y)
    conv_y = keras.layers.Activation('relu')(conv_y)

    conv_z = keras.layers.Conv1D(filters=n_feature_maps * 2, kernel_size=3, padding='same')(conv_y)
    conv_z = keras.layers.BatchNormalization()(conv_z)

    # no need to expand channels because they are equal
    shortcut_y = keras.layers.BatchNormalization()(output_block_2)

    output_block_3 = keras.layers.add([shortcut_y, conv_z])
    output_block_3 = keras.layers.Activation('relu')(output_block_3)

    # FINAL
    gap_layer = keras.layers.GlobalAveragePooling1D()(output_block_3)
    return gap_layer


# use with batch_size=12 and epochs=100
def mb_dl4tsc_encoder_mid(input_layer: keras.layers.Input):
    # conv block -1
    conv1 = keras.layers.Conv1D(filters=128, kernel_size=5, strides=1, padding='same')(input_layer)
    conv1 = InstanceNormalization()(conv1)
    conv1 = keras.layers.PReLU(shared_axes=[1])(conv1)
    conv1 = keras.layers.Dropout(rate=0.2)(conv1)
    conv1 = keras.layers.MaxPooling1D(pool_size=2)(conv1)
    # conv block -2
    conv2 = keras.layers.Conv1D(filters=256, kernel_size=11, strides=1, padding='same')(conv1)
    conv2 = InstanceNormalization()(conv2)
    conv2 = keras.layers.PReLU(shared_axes=[1])(conv2)
    conv2 = keras.layers.Dropout(rate=0.2)(conv2)
    conv2 = keras.layers.MaxPooling1D(pool_size=2)(conv2)
    # conv block -3
    conv3 = keras.layers.Conv1D(filters=512, kernel_size=21, strides=1, padding='same')(conv2)
    conv3 = InstanceNormalization()(conv3)
    conv3 = keras.layers.PReLU(shared_axes=[1])(conv3)
    conv3 = keras.layers.Dropout(rate=0.2)(conv3)
    # split for attention
    attention_data = keras.layers.Lambda(lambda x: x[:, :, :256])(conv3)
    attention_softmax = keras.layers.Lambda(lambda x: x[:, :, 256:])(conv3)
    # attention mechanism
    attention_softmax = keras.layers.Softmax()(attention_softmax)
    multiply_layer = keras.layers.Multiply()([attention_softmax, attention_data])
    # last layer
    dense_layer = keras.layers.Dense(units=256, activation='sigmoid')(multiply_layer)
    dense_layer = InstanceNormalization()(dense_layer)
    flatten_layer = keras.layers.Flatten()(dense_layer)
    return flatten_layer


def mb_inception(input_layer):
    # TODO choose which will be changable
    use_residual = True
    use_bottleneck = True
    depth = 6
    kernel_size = 41
    nb_filters = 32
    stride = 1
    activation = 'linear'

    bottleneck_size = 32
        
    x = input_layer
    input_res = input_layer

    for d in range(depth):
        if use_bottleneck and int(x.shape[-1]) > bottleneck_size:
            input_inception = keras.layers.Conv1D(filters=bottleneck_size, kernel_size=1,
                                                    padding='same', activation=activation, use_bias=False)(x)
        else:
            input_inception = x

        # kernel_size_s = [3, 5, 8, 11, 17]
        kernel_size_s = [kernel_size // (2 ** i) for i in range(3)]

        conv_list = []

        for i in range(len(kernel_size_s)):
            conv_list.append(keras.layers.Conv1D(filters=nb_filters, kernel_size=kernel_size_s[i],
                                                    strides=stride, padding='same', activation=activation, use_bias=False)(
                input_inception))

        max_pool_1 = keras.layers.MaxPool1D(pool_size=3, strides=stride, padding='same')(x)

        conv_6 = keras.layers.Conv1D(filters=nb_filters, kernel_size=1,
                                        padding='same', activation=activation, use_bias=False)(max_pool_1)

        conv_list.append(conv_6)

        x = keras.layers.Concatenate(axis=2)(conv_list)
        x = keras.layers.BatchNormalization()(x)
        x = keras.layers.Activation(activation='relu')(x)

        if use_residual and d % 3 == 2:
            shortcut_y = keras.layers.Conv1D(filters=int(x.shape[-1]), kernel_size=1,
                                                padding='same', use_bias=False)(input_res)
            shortcut_y = keras.layers.BatchNormalization()(shortcut_y)

            x = keras.layers.Add()([shortcut_y, x])
            x = keras.layers.Activation('relu')(x)
            
            input_res = x

    gap_layer = keras.layers.GlobalAveragePooling1D()(x)

    return gap_layer


def mb_tlenet(input_layer):
    conv_1 = keras.layers.Conv1D(filters=5, kernel_size=5, activation='relu', padding='same')(input_layer)
    conv_1 = keras.layers.MaxPool1D(pool_size=2)(conv_1)

    conv_2 = keras.layers.Conv1D(filters=20, kernel_size=5, activation='relu', padding='same')(conv_1)
    conv_2 = keras.layers.MaxPool1D(pool_size=4)(conv_2)

    # they did not mention the number of hidden units in the fully-connected layer
    # so we took the lenet they referenced 

    flatten_layer = keras.layers.Flatten()(conv_2)
    fully_connected_layer = keras.layers.Dense(500, activation='relu')(flatten_layer)

    return fully_connected_layer


def mb_mcnn(input_layers, pool_factor, kernel_size):
    stage_1_layers = []
    for input_layer in input_layers:
        conv_layer = keras.layers.Conv1D(filters=256, kernel_size=kernel_size, padding='same',
                                            activation='sigmoid')(input_layer)

        # should all concatenated have the same length
        pool_size = int(int(conv_layer.shape[1]) / pool_factor)
        max_layer = keras.layers.MaxPooling1D(pool_size=pool_size)(conv_layer)
        
        stage_1_layers.append(max_layer)

    concat_layer = keras.layers.Concatenate(axis=-1)(stage_1_layers)
    kernel_size = int(min(kernel_size, int(concat_layer.shape[1])))  # kernel shouldn't exceed the length
    full_conv = keras.layers.Conv1D(filters=256, kernel_size=kernel_size, padding='same',
                                    activation='sigmoid')(concat_layer)
    pool_size = int(int(full_conv.shape[1]) / pool_factor)
    full_max = keras.layers.MaxPooling1D(pool_size=pool_size)(full_conv)
    full_max = keras.layers.Flatten()(full_max)
    fully_connected = keras.layers.Dense(units=256, activation='sigmoid')(full_max)

    return fully_connected


def mb_mcdcnn(input_layers, n_t, n_vars):
    # TODO use padding var in mb_dl4tsc to choose independent of n_t?
    padding = 'valid'
    if n_t < 60:  # for ItalyPowerOndemand
        padding = 'same'

    conv2_layers = []
    for input_layer in input_layers:
        conv1_layer = keras.layers.Conv1D(filters=8, kernel_size=5, activation='relu', padding=padding)(input_layer)
        conv1_layer = keras.layers.MaxPooling1D(pool_size=2)(conv1_layer)

        conv2_layer = keras.layers.Conv1D(filters=8, kernel_size=5, activation='relu', padding=padding)(conv1_layer)
        conv2_layer = keras.layers.MaxPooling1D(pool_size=2)(conv2_layer)
        conv2_layer = keras.layers.Flatten()(conv2_layer)

        conv2_layers.append(conv2_layer)

    if n_vars == 1:
        # to work with univariate time series
        concat_layer = conv2_layers[0]
    else:
        concat_layer = keras.layers.Concatenate(axis=-1)(conv2_layers)

    fully_connected = keras.layers.Dense(units=732, activation='relu')(concat_layer)

    return fully_connected


def mb_ad_mid(input_layer: keras.layers.Input):
    flatten_layer = keras.layers.Flatten()(input_layer)
    encode_layer = keras.layers.Dense(50, activation='relu')(flatten_layer)

    return encode_layer


def mb_ad_lower(architecture: ModelArchitecture, prev_layer, nb_classes, input_shape):
    metrics = ['accuracy', 'mse']

    loss = 'binary_crossentropy'

    learning_rate = 0.001

    optimizer = keras.optimizers.Adam(learning_rate)

    hidden_size = np.prod(np.array(input_shape))
    decode_layer = keras.layers.Dense(hidden_size, activation='sigmoid')(prev_layer)
    output_layer = keras.layers.Reshape(input_shape)(decode_layer)

    return output_layer, (optimizer, loss, metrics)


# TODO name doesn't fit purpose anymore
def mb_dl4tsc(architecture: ModelArchitecture, input_shape: tuple,
              pool_factor=None, kernel_size=None):
    if architecture == ModelArchitecture.MCNN:
        input_layer = []
        for shape in input_shape:
            input_layer.append(keras.layers.Input(shape))
        
        # self.pool_factors = [2, 3, 5]
        mid = mb_mcnn(input_layer, pool_factor, kernel_size)
    elif architecture == ModelArchitecture.MCDCNN:
        n_t = input_shape[0]
        n_vars = input_shape[1]
        
        input_layer = []
        for _ in range(n_vars):
            input_layer.append(keras.layers.Input((n_t, 1)))
        
        mid = mb_mcdcnn(input_layer, n_t, n_vars)
    else:
        input_layer = keras.layers.Input(input_shape)
    
        if architecture == ModelArchitecture.CNN_VALID:
            mid = _mb_dl4tsc_cnn_mid(input_layer, "valid")

        if architecture == ModelArchitecture.CNN_SAME:
            mid = _mb_dl4tsc_cnn_mid(input_layer, "same")

        if architecture == ModelArchitecture.FCN:
            mid = mb_dl4tsc_fcn_mid(input_layer)

        if architecture == ModelArchitecture.MLP:
            mid = mb_dl4tsc_mlp_mid(input_layer)

        if architecture == ModelArchitecture.RESNET:
            mid = mb_dl4tsc_resnet_mid(input_layer)

        if architecture == ModelArchitecture.ENCODER:
            mid = mb_dl4tsc_encoder_mid(input_layer)
        
        if architecture == ModelArchitecture.INCEPTION:
            mid = mb_inception(input_layer)

        if architecture == ModelArchitecture.TLENET:
            mid = mb_tlenet(input_layer)

        if architecture == ModelArchitecture.TWIESN:
            raise NotImplementedError("TWIESN architecture is not implemented yet.")

        if architecture == ModelArchitecture.AD:
            mid = mb_ad_mid(input_layer)

    return input_layer, mid


def mb_dl4tsc_lower(architecture: ModelArchitecture, prev_layer, nb_classes):
    metrics = ['accuracy']
    if architecture == ModelArchitecture.AD:
        metrics = ['accuracy', 'mse']

    loss = 'categorical_crossentropy'
    if architecture == ModelArchitecture.AD:
        loss = 'binary_crossentropy'

    activation = 'softmax'
    if architecture == ModelArchitecture.CNN_VALID \
    or architecture == ModelArchitecture.CNN_SAME:
        loss = 'mean_squared_error'
        activation = 'sigmoid'
    
    learning_rate = 0.001
    if architecture == ModelArchitecture.ENCODER:
        learning_rate = 0.00001
    if architecture == ModelArchitecture.MCNN:
        learning_rate = 0.1
    if architecture == ModelArchitecture.MCDCNN \
    or architecture == ModelArchitecture.TLENET:
        learning_rate = 0.01
    
    optimizer = keras.optimizers.Adam(learning_rate)
    if architecture == ModelArchitecture.MLP:
        optimizer = keras.optimizers.Adadelta()
    if architecture == ModelArchitecture.MCDCNN:
        optimizer = keras.optimizers.SGD(learning_rate, momentum=0.9, decay=0.0005)
    if architecture == ModelArchitecture.TLENET:
        optimizer = keras.optimizers.Adam(learning_rate, decay=0.005)
    if architecture == ModelArchitecture.INCEPTION:
        optimizer = keras.optimizers.Adam(learning_rate)
    
    output_layer = prev_layer
    if architecture != ModelArchitecture.AD:
        output_layer = keras.layers.Dense(nb_classes, activation=activation)(prev_layer)

    return output_layer, (optimizer, loss, metrics)

def mb_dl4cl_lower(architecture: ModelArchitecture, prev_layer, nb_classes):
    metrics = ['accuracy']

    optimizer="RMSprop"
    
    loss = 'contrastive_loss'
    
    activation = 'sigmoid'
    merge_layer = tf.linalg.norm((prev_layer(input) - prev_layer(input)), axis=1, keepdims=True)
    normal_layer = tf.keras.layers.BatchNormalization()(merge_layer)
    output_layer = tf.keras.layers.Dense(1, activation=activation)(normal_layer)
    return output_layer, (optimizer, loss, metrics)

def mb_dl4tl_lower(architecture: ModelArchitecture, prev_layer, nb_classes):
    pass
    metrics = ['accuracy']
    optimizer = f"{keras.optimizers.Adam(0.0001)}"
    loss = TripletSemiHardLoss()
    t0 = prev_layer #anchor
    t1 = prev_layer #positive
    t2 = prev_layer #negative
    ap_distance = tf.keras.layers.Layer(tf.reduce_sum(tf.square(t0 - t1), -1))
    an_distance = tf.keras.layers.Layer(tf.reduce_sum(tf.square(t0 - t2), -1))

def mb_bv_lower(architecture: ModelArchitecture, prev_layer, nb_classes):
    loss = 'binary_crossentropy'
    optimizer = keras.optimizers.Adam()
    metrics = ['accuracy']
    output_layer = keras.layers.Dense(1, activation='sigmoid')(prev_layer)
    return output_layer, (optimizer, loss, metrics)



import unittest
import numpy as np
import keras
from tensorflow.keras.datasets import mnist
from tensorflow.keras import backend as K
from sklearn.metrics import confusion_matrix


class TestSiameseNetwork(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        # Hyperparameters
        cls.epochs = 10
        cls.batch_size = 16
        cls.margin = 1  # Margin for contrastive loss

        # Load and preprocess MNIST data
        (x_train_val, y_train_val), (x_test, y_test) = mnist.load_data()

        # Add channel dimension and normalize
        x_train_val = x_train_val.astype("float32") / 255.0
        x_train_val = np.expand_dims(x_train_val, -1)  # Add channel dimension

        x_test = x_test.astype("float32") / 255.0
        x_test = np.expand_dims(x_test, -1)  # Add channel dimension

        # Split train/validation
        cls.x_train, cls.x_val = x_train_val[:30000], x_train_val[30000:]
        cls.y_train, cls.y_val = y_train_val[:30000], y_train_val[30000:]
        cls.x_test, cls.y_test = x_test, y_test

        # Create pairs
        cls.pairs_train, cls.labels_train = cls.make_pairs(cls.x_train, cls.y_train)
        cls.pairs_val, cls.labels_val = cls.make_pairs(cls.x_val, cls.y_val)
        cls.pairs_test, cls.labels_test = cls.make_pairs(cls.x_test, cls.y_test)

    @staticmethod
    def make_pairs(x, y):
        num_classes = max(y) + 1
        digit_indices = [np.where(y == i)[0] for i in range(num_classes)]

        pairs = []
        labels = []

        for idx1 in range(len(x)):
            x1 = x[idx1]
            label1 = y[idx1]

            # Positive pair
            idx2 = np.random.choice(digit_indices[label1])
            x2 = x[idx2]
            pairs += [[x1, x2]]
            labels += [0]  # 0 means same class (positive pair)

            # Negative pair
            label2 = np.random.randint(0, num_classes)
            while label2 == label1:
                label2 = np.random.randint(0, num_classes)

            idx2 = np.random.choice(digit_indices[label2])
            x2 = x[idx2]
            pairs += [[x1, x2]]
            labels += [1]  # 1 means different class (negative pair)

        return np.array(pairs), np.array(labels).astype("float32")

    def test_siamese_network(self):
        # Build model
        siamese_model = self.create_siamese_model()

        # Use a smaller learning rate to prevent NaN loss
        optimizer = keras.optimizers.RMSprop(learning_rate=0.0005)

        siamese_model.compile(
            loss=self.contrastive_loss(margin=self.margin),
            optimizer=optimizer,
            metrics=["accuracy"]
        )

        # Prepare data
        x_train_1, x_train_2 = self.pairs_train[:, 0], self.pairs_train[:, 1]
        x_val_1, x_val_2 = self.pairs_val[:, 0], self.pairs_val[:, 1]

        # Train
        history = siamese_model.fit(
            [x_train_1, x_train_2],
            self.labels_train,
            validation_data=([x_val_1, x_val_2], self.labels_val),
            batch_size=self.batch_size,
            epochs=self.epochs
        )

        # Evaluate
        x_test_1, x_test_2 = self.pairs_test[:, 0], self.pairs_test[:, 1]
        results = siamese_model.evaluate([x_test_1, x_test_2], self.labels_test)
        print("Test loss, test acc:", results)

        # Generate predictions
        predictions = siamese_model.predict([x_test_1, x_test_2])
        y_pred = (predictions > 0.5).astype("int32")

        # Generate confusion matrix
        cm = confusion_matrix(self.labels_test, y_pred)
        print("Confusion Matrix:\n", cm)

    def create_siamese_model(self):
        # Embedding network with L2 regularization
        input = keras.layers.Input((28, 28, 1))

        # Add batch normalization at the input
        x = keras.layers.BatchNormalization()(input)

        x = keras.layers.Conv2D(4, (5, 5), activation="relu",
                                kernel_regularizer=keras.regularizers.l2(1e-4))(x)
        x = keras.layers.AveragePooling2D(pool_size=(2, 2))(x)

        x = keras.layers.Conv2D(16, (5, 5), activation="relu",
                                kernel_regularizer=keras.regularizers.l2(1e-4))(x)
        x = keras.layers.AveragePooling2D(pool_size=(2, 2))(x)

        x = keras.layers.Flatten()(x)

        # Change to ReLU and add regularization
        x = keras.layers.Dense(10, activation="relu",
                               kernel_regularizer=keras.regularizers.l2(1e-4))(x)

        # Add a final batch normalization
        x = keras.layers.BatchNormalization()(x)

        embedding_network = keras.Model(input, x)

        # Siamese network
        input_1 = keras.layers.Input((28, 28, 1))
        input_2 = keras.layers.Input((28, 28, 1))
        tower_1 = embedding_network(input_1)
        tower_2 = embedding_network(input_2)

        # Add small epsilon to prevent numerical instability
        merge_layer = keras.layers.Lambda(
            lambda vects: K.sqrt(K.sum(K.square(vects[0] - vects[1]), axis=1, keepdims=True) + 1e-10)
        )([tower_1, tower_2])

        # Remove the batch normalization after distance calculation
        output_layer = keras.layers.Dense(1, activation="sigmoid")(merge_layer)
        return keras.Model(inputs=[input_1, input_2], outputs=output_layer)

    def contrastive_loss(self, margin=1):
        def loss(y_true, y_pred):
            # Add small epsilon to avoid numerical instability
            square_pred = K.square(y_pred)
            margin_square = K.square(K.maximum(margin - y_pred, 0))
            return K.mean((1 - y_true) * square_pred + y_true * margin_square)

        return loss


if __name__ == '__main__':
    unittest.main()


from abc import ABC, abstractmethod

import numpy as np
from sklearn.ensemble import RandomForestRegressor
from tensorflow import keras
from tensorflow.keras import Sequential
from tensorflow.keras import layers


class KerasBaseDescription(ABC):

    @abstractmethod
    def build_model(self, input_shape) -> keras.models.Model:
        pass

    @abstractmethod
    def callbacks(self) -> [keras.callbacks.Callback]:
        pass


class SequentialAutoEncoder(KerasBaseDescription):

    def build_model(self, input_shape):
        hidden_size = np.prod(np.array(input_shape))

        model = Sequential()
        model.add(layers.Flatten(input_shape=input_shape))
        model.add(layers.Dense(50, activation='relu'))
        model.add(layers.Dense(hidden_size, activation='sigmoid'))
        model.add(layers.Reshape(input_shape))
        model.build(input_shape=input_shape)
        model.compile(optimizer='adam', loss='binary_crossentropy', metrics=['acc', 'mse'])
        return model

    def callbacks(self) -> [keras.callbacks.Callback]:
        return [keras.callbacks.EarlyStopping(monitor="mse",
                                              patience=25,
                                              mode="min",
                                              restore_best_weights=True)]



import unittest

from useckit.v2.dataset import Dataset
from useckit.v2.keras_model_builder import ModelBuilder
from useckit.v2.evaluator import Paradigm
from useckit.v2.model_descriptions import MODEL_DESCRIPTIONS
from useckit.v2.utils.example_dataset import get_chi2021_dataset


class TestKerasModelBuilder(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        # Use feature vectors so that sklearn models can be trained easily
        cls.ds: Dataset = get_chi2021_dataset(feature_vector=True)

    def _build_validate_train(self, p: Paradigm, model_desc_key: str):
        print("Testing paradigm", p, "/", model_desc_key)

        mb = ModelBuilder(MODEL_DESCRIPTIONS)

        input_shape = self.ds.get_shape()
        num_classes = self.ds.get_unique_labels().shape[0]

        # Build steps
        if p == Paradigm.CSC:
            build_result = mb.build_multiclass_classification(
                arch_key=model_desc_key,
                input_shape=input_shape,
                num_classes=num_classes,
            )
        elif p == Paradigm.DL:
            build_result1 = mb.build_metriclearning_triplet(model_desc_key, input_shape=input_shape)
            build_result2 = mb.build_metriclearning_contrastive(model_desc_key, input_shape=input_shape)
        elif p == Paradigm.BV:
            build_result = mb.build_binary_verification_classifier(model_desc_key, input_shape=input_shape)
        elif p == Paradigm.AD:
            build_result = mb.build_autoencoder_oneclass(model_desc_key, input_shape=input_shape)
        else:
            raise ValueError(f"Unknown paradigm: {p}")

        # Validation Steps
        if p == Paradigm.CSC:
            build_result["model"].summary()
            assert build_result["model"].layers[-1].units == len(self.ds.get_unique_enrollment_labels()), \
                f"Expected {len(self.ds.get_unique_enrollment_labels())} units in the last layer, but got {build_result['model'].layers[-1].units} units."
            assert build_result["model"].loss == "categorical_crossentropy", \
                f"Expected categorical_crossentropy loss, but got {build_result['model'].loss}."
        elif p == Paradigm.DL:
            for build_result in [build_result1, build_result2]:
                build_result["model"].summary()
                from keras.src.layers import Lambda
                assert isinstance(build_result["model"].layers[-1], Lambda), \
                    f"Expected Lambda in the last layer, but got {type(build_result['model'].layers[-1])} instead."
        elif p == Paradigm.BV:
            build_result["model"].summary()
            assert build_result["model"].layers[-1].units == 1, \
                f"Expected 1 unit in the last layer, but got {build_result['model'].layers[-1].units} units."
            assert build_result["model"].loss == "binary_crossentropy", \
                f"Expected binary_crossentropy loss, but got {build_result['model'].loss}."
        elif p == Paradigm.AD:
            build_result["model"].summary()
            from keras.src.layers import Conv1D
            assert isinstance(build_result["model"].layers[-1], Conv1D), \
                f"Expected Conv1D in the last layer, but got {type(build_result['model'].layers[-1])} instead."

        # Training Steps
        verbose=1
        epochs=10
        if p == Paradigm.CSC:
            model, callbacks, metadata = build_result["model"], build_result["callbacks"], build_result["metadata"]
            model.fit(x=self.ds.get_trainset_data(),
                      y=self.ds.get_trainset_labels(),
                      callbacks=callbacks,
                      verbose=verbose,
                      epochs=epochs)
        elif p == Paradigm.DL:
            for build_result in [build_result1, build_result2]:
                model, callbacks, metadata = build_result["model"], build_result["callbacks"], build_result["metadata"]
                model.fit(x=self.ds.get_trainset_data(),
                          y=self.ds.get_trainset_labels(),
                          callbacks=callbacks,
                          verbose=verbose,
                          epochs=epochs)
        elif p == Paradigm.BV:
            model, callbacks, metadata = build_result["model"], build_result["callbacks"], build_result["metadata"]
            model.fit(x=self.ds.get_trainset_data(),
                      y=self.ds.get_trainset_labels(),
                      callbacks=callbacks,
                      verbose=verbose,
                      epochs=epochs)

        elif p == Paradigm.AD:
            model, callbacks, metadata = build_result["model"], build_result["callbacks"], build_result["metadata"]
            model.fit(x=self.ds.get_trainset_data(),
                      y=self.ds.get_trainset_labels(),
                      callbacks=callbacks,
                      verbose=verbose,
                      epochs=epochs)

        print("End.")


    # ---- Hardcoded test cases: CSC ----
    def test_csc_cnn_valid(self):
        self._build_validate_train(Paradigm.CSC, 'CNN_VALID')

    def test_csc_cnn_same(self):
        self._build_validate_train(Paradigm.CSC, 'CNN_SAME')

    def test_csc_fcn(self):
        self._build_validate_train(Paradigm.CSC, 'FCN')

    def test_csc_mlp(self):
        self._build_validate_train(Paradigm.CSC, 'MLP')

    def test_csc_resnet(self):
        self._build_validate_train(Paradigm.CSC, 'RESNET')

    def test_csc_encoder(self):
        self._build_validate_train(Paradigm.CSC, 'ENCODER')

    def test_csc_inception(self):
        self._build_validate_train(Paradigm.CSC, 'INCEPTION')

    def test_csc_tlenet(self):
        self._build_validate_train(Paradigm.CSC, 'TLENET')

    def test_csc_mcdcnn(self):
        self._build_validate_train(Paradigm.CSC, 'MCDCNN')

    def test_csc_mcnn(self):
        self._build_validate_train(Paradigm.CSC, 'MCNN')

    def test_csc_twiesn(self):
        self._build_validate_train(Paradigm.CSC, 'TWIESN')

    # ---- Hardcoded test cases: DL ----
    def test_dl_cnn_valid(self):
        self._build_validate_train(Paradigm.DL, 'CNN_VALID')

    def test_dl_cnn_same(self):
        self._build_validate_train(Paradigm.DL, 'CNN_SAME')

    def test_dl_fcn(self):
        self._build_validate_train(Paradigm.DL, 'FCN')

    def test_dl_mlp(self):
        self._build_validate_train(Paradigm.DL, 'MLP')

    def test_dl_resnet(self):
        self._build_validate_train(Paradigm.DL, 'RESNET')

    def test_dl_encoder(self):
        self._build_validate_train(Paradigm.DL, 'ENCODER')

    def test_dl_inception(self):
        self._build_validate_train(Paradigm.DL, 'INCEPTION')

    def test_dl_tlenet(self):
        self._build_validate_train(Paradigm.DL, 'TLENET')

    def test_dl_mcdcnn(self):
        self._build_validate_train(Paradigm.DL, 'MCDCNN')

    def test_dl_mcnn(self):
        self._build_validate_train(Paradigm.DL, 'MCNN')

    def test_dl_twiesn(self):
        self._build_validate_train(Paradigm.DL, 'TWIESN')

    # ---- Hardcoded test cases: BV ----
    def test_bv_cnn_valid(self):
        self._build_validate_train(Paradigm.BV, 'CNN_VALID')

    def test_bv_cnn_same(self):
        self._build_validate_train(Paradigm.BV, 'CNN_SAME')

    def test_bv_fcn(self):
        self._build_validate_train(Paradigm.BV, 'FCN')

    def test_bv_mlp(self):
        self._build_validate_train(Paradigm.BV, 'MLP')

    def test_bv_resnet(self):
        self._build_validate_train(Paradigm.BV, 'RESNET')

    def test_bv_encoder(self):
        self._build_validate_train(Paradigm.BV, 'ENCODER')

    def test_bv_inception(self):
        self._build_validate_train(Paradigm.BV, 'INCEPTION')

    def test_bv_tlenet(self):
        self._build_validate_train(Paradigm.BV, 'TLENET')

    def test_bv_mcdcnn(self):
        self._build_validate_train(Paradigm.BV, 'MCDCNN')

    def test_bv_mcnn(self):
        self._build_validate_train(Paradigm.BV, 'MCNN')

    def test_bv_twiesn(self):
        self._build_validate_train(Paradigm.BV, 'TWIESN')

    # ---- Hardcoded test cases: AD ----
    def test_ad_cnn_valid(self):
        self._build_validate_train(Paradigm.AD, 'CNN_VALID')

    def test_ad_cnn_same(self):
        self._build_validate_train(Paradigm.AD, 'CNN_SAME')

    def test_ad_fcn(self):
        self._build_validate_train(Paradigm.AD, 'FCN')

    def test_ad_mlp(self):
        self._build_validate_train(Paradigm.AD, 'MLP')

    def test_ad_resnet(self):
        self._build_validate_train(Paradigm.AD, 'RESNET')

    def test_ad_encoder(self):
        self._build_validate_train(Paradigm.AD, 'ENCODER')

    def test_ad_inception(self):
        self._build_validate_train(Paradigm.AD, 'INCEPTION')

    def test_ad_tlenet(self):
        self._build_validate_train(Paradigm.AD, 'TLENET')

    def test_ad_mcdcnn(self):
        self._build_validate_train(Paradigm.AD, 'MCDCNN')

    def test_ad_mcnn(self):
        self._build_validate_train(Paradigm.AD, 'MCNN')

    def test_ad_twiesn(self):
        self._build_validate_train(Paradigm.AD, 'TWIESN')



if __name__ == '__main__':
    unittest.main()



import re
import unittest
import os
import shutil
import urllib.request
import hashlib
import subprocess
from glob import glob
from statistics import median

import numpy as np
import pandas as pd
import tensorflow as tf
from joblib import Parallel, delayed
from keras import Sequential
from keras.src.layers import Dense
from keras.src.optimizers import Adam
from keras.src.utils import pad_sequences
from sklearn.metrics import confusion_matrix
from sklearn.preprocessing import LabelBinarizer

import useckit

class TestModelBuilder_DIST(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        self.unittest_execution_counter = 1
        # make train pairs
        self.pairs_train, self.labels_train = self.make_pairs(x_train, y_train)

        # make validation pairs
        self.pairs_val, self.labels_val = self.make_pairs(x_val, y_val)

        # make test pairs
        self.pairs_test, self.labels_test = self.make_pairs(x_test, y_test)

    @classmethod
    def prepare_mnist_dataset(self):
        # Set custom save path
        save_path = "../mnist_data"
        os.makedirs(save_path, exist_ok=True)

        # Load MNIST dataset
        (x_train, y_train), (x_test, y_test) = tf.keras.datasets.mnist.load_data()

        # Save dataset to custom location
        np.save(os.path.join(save_path, "x_train.npy"), x_train)
        np.save(os.path.join(save_path, "y_train.npy"), y_train)
        np.save(os.path.join(save_path, "x_test.npy"), x_test)
        np.save(os.path.join(save_path, "y_test.npy"), y_test)

        print("MNIST dataset saved successfully!")

    def make_pairs(x, y):
        """Creates a tuple containing image pairs with corresponding label.

        Arguments:
            x: List containing images, each index in this list corresponds to one image.
            y: List containing labels, each label with datatype of `int`.

        Returns:
            Tuple containing two numpy arrays as (pairs_of_samples, labels),
            where pairs_of_samples' shape is (2len(x), 2,n_features_dims) and
            labels are a binary array of shape (2len(x)).
        """

        num_classes = max(y) + 1
        digit_indices = [np.where(y == i)[0] for i in range(num_classes)]

        pairs = []
        labels = []

        for idx1 in range(len(x)):
            # add a matching example
            x1 = x[idx1]
            label1 = y[idx1]
            idx2 = random.choice(digit_indices[label1])
            x2 = x[idx2]

            pairs += [[x1, x2]]
            labels += [0]

            # add a non-matching example
            label2 = random.randint(0, num_classes - 1)
            while label2 == label1:
                label2 = random.randint(0, num_classes - 1)

            idx2 = random.choice(digit_indices[label2])
            x2 = x[idx2]

            pairs += [[x1, x2]]
            labels += [1]

        return np.array(pairs), np.array(labels).astype("float32")

if __name__ == '__main__':
    unittest.main()


import numpy as np

from ._equal_error_thresholding_method import EqualErrorThresholding
from ._sample_broadcasting import _create_all_distance_pairs
from .distance_evaluation_method_base import DistanceBaseEvaluationMethod, BaseThresholdingMethod
from ..prediction_models.distance_prediction_model_base import DistanceBasePredictionModel
from ....evaluation.identification_with_reject import IdentificationOrRejectModel, \
    perform_identification_or_reject_evaluation
from ....util.dataset import Dataset
from ....util.utils import contrastive_make_pairs


class DistanceIdentificationOrRejectModel(IdentificationOrRejectModel):

    def __init__(self,
                 distance_metric: DistanceBasePredictionModel,
                 enrollment_samples: np.ndarray,
                 enrollment_labels: np.ndarray,
                 rejection_threshold: float,
                 dataset: Dataset,
                 tradeoff_computation_speed_for_memory: bool):
        self.distance_metric = distance_metric
        self.enrolment_samples = enrollment_samples
        self.enrollment_labels = enrollment_labels
        self.rejection_threshold = rejection_threshold
        self.dataset = dataset
        self.tradeoff_computation_speed_for_memory = tradeoff_computation_speed_for_memory

    def identify_or_reject(self, samples: np.ndarray):
        result = np.zeros(shape=(len(samples),), dtype=int)

        distances, sample_indices = _create_all_distance_pairs(samples, self.enrolment_samples, self.distance_metric,
                                                               self.tradeoff_computation_speed_for_memory)

        for index, sample in enumerate(samples):
            index_mask = sample_indices == index
            distances_to_enrollment = distances[index_mask]  # These are all the predicted distances of this
            # loop-repetition's sample to all enrolment samples. They are ordered by the order that the
            # self.enrolment_samples array imposes

            min_median_distance = 1e1000  # just a stupidly large number
            min_median_label = 0  # just any number
            for enrol_label in np.unique(self.enrollment_labels):
                distances_to_enrol_label = distances_to_enrollment[self.enrollment_labels == enrol_label]  # These are
                # all the distances to the enrolment samples belonging to enrol_label
                if len(distances_to_enrol_label) <= 2:
                    median = min(distances_to_enrol_label)  # don't find the value closest to the median for just two
                    # values, but pick the smaller one
                else:
                    median_dist_index = \
                        np.nonzero(
                            distances_to_enrol_label == np.percentile(distances_to_enrol_label, 50,
                                                                      method='closest_observation')
                        )[0][0]  # This abomination finds the index of the closest value in the array to the median
                    median = distances_to_enrol_label[median_dist_index]  # median of distances to the enrolment samples
                    # belonging to enrol_label
                if median < min_median_distance:
                    min_median_distance = median
                    min_median_label = enrol_label
            if min_median_distance >= self.rejection_threshold:
                result[index] = -1
            else:
                result[index] = min_median_label
        return self.dataset.reverse_label_transform(result)


class IdentificationWithReject(DistanceBaseEvaluationMethod):

    def __init__(self,
                 output_dir: str = "evaluation_identification_with_reject",
                 threshold_method: BaseThresholdingMethod = None,
                 tradeoff_computation_speed_for_memory: bool = True):
        super().__init__(output_dir)
        self.threshold_method = threshold_method
        self.tradeoff_computation_speed_for_memory = tradeoff_computation_speed_for_memory

    def evaluate(self, dataset: Dataset, prediction_model: DistanceBasePredictionModel, **kwargs):
        if self.threshold_method is None:
            self.threshold_method = EqualErrorThresholding(prediction_model, contrastive_make_pairs, self.output_dir)

        threshold = self.threshold_method.compute_threshold(dataset.testset_enrollment_data,
                                                            dataset.testset_enrollment_labels)
        ident_or_reject_model = DistanceIdentificationOrRejectModel(prediction_model,
                                                                    dataset.testset_enrollment_data,
                                                                    dataset.testset_enrollment_labels,
                                                                    threshold,
                                                                    dataset, self.tradeoff_computation_speed_for_memory)
        perform_identification_or_reject_evaluation(ident_or_reject_model, dataset, self.output_dir)



# ATTRIBUTION: this file consolidates model descriptions originating from
# https://github.com/hfawaz/dl-4-tsc
# It was published on github.com by Ismail Fawaz, Hassan and Forestier, Germain
# and Weber, Jonathan and Idoumghar, Lhassane and Muller, Pierre-Alain
# for their work "Deep learning for time series classification: a review",
# in Data Mining and Knowledge Discovery.
# LICENSE is GNU GENERAL PUBLIC LICENSE version 3 (29 June 2007).


"""
Centralized model descriptions for Time Series Classification (TSC) models (v2).

This module collates callable model descriptions for the eleven TSC models:
- CNN (valid), CNN (same), FCN, MLP, RESNET, ENCODER
- INCEPTION, TLENET, MCDCNN, MCNN, TWIESN

Notes:
- Keras-based descriptions return a tuple: (compiled_keras_model, [callbacks]).
- Some legacy models (e.g., TWIESN) are not Keras graphs. For these, we expose
  a descriptor callable that raises NotImplementedError for Keras usage and
  provide a reference to the legacy class in the docstring for integration.
"""

import gc
import os
import pickle
import time
import numpy as np
from scipy import sparse
from scipy.sparse import linalg as slinalg
from sklearn.linear_model import Ridge
from typing import Callable, List, Tuple

import tensorflow.keras as keras



def _dl4tsc_cnn_constructor_helper(padding: str, input_layer, nb_classes) -> Tuple["keras.Model", List[Callable]]:  
    conv1 = keras.layers.Conv1D(filters=6, kernel_size=7, padding=padding, activation='sigmoid')(input_layer)
    conv1 = keras.layers.AveragePooling1D(pool_size=3)(conv1)

    conv2 = keras.layers.Conv1D(filters=12, kernel_size=7, padding=padding, activation='sigmoid')(conv1)
    conv2 = keras.layers.AveragePooling1D(pool_size=3)(conv2)

    flatten_layer = keras.layers.Flatten()(conv2)

    output_layer = keras.layers.Dense(units=nb_classes, activation='sigmoid')(flatten_layer)

    model = keras.models.Model(inputs=input_layer, outputs=output_layer)

    model.compile(loss='mean_squared_error', optimizer=keras.optimizers.Adam(), metrics=['accuracy'])

    return model, []


def dl4tsc_cnn_padding_valid(input_layer, nb_classes) -> Tuple["keras.Model", List[Callable]]:  
    return _dl4tsc_cnn_constructor_helper('valid', input_layer, nb_classes)


def dl4tsc_cnn_padding_same(input_layer, nb_classes) -> Tuple["keras.Model", List[Callable]]:  
    return _dl4tsc_cnn_constructor_helper('same', input_layer, nb_classes)


def dl4tsc_fcn(input_layer, nb_classes) -> Tuple["keras.Model", List[Callable]]:  
    conv1 = keras.layers.Conv1D(filters=128, kernel_size=8, padding='same')(input_layer)
    conv1 = keras.layers.BatchNormalization()(conv1)
    conv1 = keras.layers.Activation(activation='relu')(conv1)

    conv2 = keras.layers.Conv1D(filters=256, kernel_size=5, padding='same')(conv1)
    conv2 = keras.layers.BatchNormalization()(conv2)
    conv2 = keras.layers.Activation('relu')(conv2)

    conv3 = keras.layers.Conv1D(128, kernel_size=3, padding='same')(conv2)
    conv3 = keras.layers.BatchNormalization()(conv3)
    conv3 = keras.layers.Activation('relu')(conv3)

    gap_layer = keras.layers.GlobalAveragePooling1D()(conv3)

    output_layer = keras.layers.Dense(nb_classes, activation='softmax')(gap_layer)

    model = keras.models.Model(inputs=input_layer, outputs=output_layer)

    model.compile(loss='categorical_crossentropy', optimizer=keras.optimizers.Adam(), metrics=['accuracy'])

    reduce_lr = keras.callbacks.ReduceLROnPlateau(monitor='loss', factor=0.5, patience=50, min_lr=0.0001)
    return model, [reduce_lr]


def dl4tsc_mlp(input_layer, nb_classes) -> Tuple["keras.Model", List[Callable]]:  
    input_layer_flattened = keras.layers.Flatten()(input_layer)

    layer_1 = keras.layers.Dropout(0.1)(input_layer_flattened)
    layer_1 = keras.layers.Dense(500, activation='relu')(layer_1)

    layer_2 = keras.layers.Dropout(0.2)(layer_1)
    layer_2 = keras.layers.Dense(500, activation='relu')(layer_2)

    layer_3 = keras.layers.Dropout(0.2)(layer_2)
    layer_3 = keras.layers.Dense(500, activation='relu')(layer_3)

    output_layer = keras.layers.Dropout(0.3)(layer_3)
    output_layer = keras.layers.Dense(nb_classes, activation='softmax')(output_layer)

    model = keras.models.Model(inputs=input_layer, outputs=output_layer)

    model.compile(loss='categorical_crossentropy', optimizer=keras.optimizers.Adadelta(), metrics=['accuracy'])

    reduce_lr = keras.callbacks.ReduceLROnPlateau(monitor='loss', factor=0.5, patience=200, min_lr=0.1)

    return model, [reduce_lr]


def dl4tsc_resnet(input_layer, nb_classes) -> Tuple["keras.Model", List[Callable]]:  
    n_feature_maps = 64

    conv_x = keras.layers.Conv1D(filters=n_feature_maps, kernel_size=8, padding='same')(input_layer)
    conv_x = keras.layers.BatchNormalization()(conv_x)
    conv_x = keras.layers.Activation('relu')(conv_x)

    conv_y = keras.layers.Conv1D(filters=n_feature_maps, kernel_size=5, padding='same')(conv_x)
    conv_y = keras.layers.BatchNormalization()(conv_y)
    conv_y = keras.layers.Activation('relu')(conv_y)

    conv_z = keras.layers.Conv1D(filters=n_feature_maps, kernel_size=3, padding='same')(conv_y)
    conv_z = keras.layers.BatchNormalization()(conv_z)

    shortcut_y = keras.layers.Conv1D(filters=n_feature_maps, kernel_size=1, padding='same')(input_layer)
    shortcut_y = keras.layers.BatchNormalization()(shortcut_y)

    output_block_1 = keras.layers.add([shortcut_y, conv_z])
    output_block_1 = keras.layers.Activation('relu')(output_block_1)

    conv_x = keras.layers.Conv1D(filters=n_feature_maps * 2, kernel_size=8, padding='same')(output_block_1)
    conv_x = keras.layers.BatchNormalization()(conv_x)
    conv_x = keras.layers.Activation('relu')(conv_x)

    conv_y = keras.layers.Conv1D(filters=n_feature_maps * 2, kernel_size=5, padding='same')(conv_x)
    conv_y = keras.layers.BatchNormalization()(conv_y)
    conv_y = keras.layers.Activation('relu')(conv_y)

    conv_z = keras.layers.Conv1D(filters=n_feature_maps * 2, kernel_size=3, padding='same')(conv_y)
    conv_z = keras.layers.BatchNormalization()(conv_z)

    shortcut_y = keras.layers.Conv1D(filters=n_feature_maps * 2, kernel_size=1, padding='same')(output_block_1)
    shortcut_y = keras.layers.BatchNormalization()(shortcut_y)

    output_block_2 = keras.layers.add([shortcut_y, conv_z])
    output_block_2 = keras.layers.Activation('relu')(output_block_2)

    conv_x = keras.layers.Conv1D(filters=n_feature_maps * 2, kernel_size=8, padding='same')(output_block_2)
    conv_x = keras.layers.BatchNormalization()(conv_x)
    conv_x = keras.layers.Activation('relu')(conv_x)

    conv_y = keras.layers.Conv1D(filters=n_feature_maps * 2, kernel_size=5, padding='same')(conv_x)
    conv_y = keras.layers.BatchNormalization()(conv_y)
    conv_y = keras.layers.Activation('relu')(conv_y)

    conv_z = keras.layers.Conv1D(filters=n_feature_maps * 2, kernel_size=3, padding='same')(conv_y)
    conv_z = keras.layers.BatchNormalization()(conv_z)

    shortcut_y = keras.layers.BatchNormalization()(output_block_2)

    output_block_3 = keras.layers.add([shortcut_y, conv_z])
    output_block_3 = keras.layers.Activation('relu')(output_block_3)

    gap_layer = keras.layers.GlobalAveragePooling1D()(output_block_3)

    output_layer = keras.layers.Dense(nb_classes, activation='softmax')(gap_layer)

    model = keras.models.Model(inputs=input_layer, outputs=output_layer)

    model.compile(loss='categorical_crossentropy', optimizer=keras.optimizers.Adam(), metrics=['accuracy'])

    reduce_lr = keras.callbacks.ReduceLROnPlateau(monitor='loss', factor=0.5, patience=50, min_lr=0.0001)

    return model, [reduce_lr]


def dl4tsc_encoder(input_layer, nb_classes) -> Tuple["keras.Model", List[Callable]]:  
    from useckit.paradigms.time_series_classification.prediction_models.instance_normalization_layer import InstanceNormalization
    conv1 = keras.layers.Conv1D(filters=128, kernel_size=5, strides=1, padding='same')(input_layer)
    conv1 = InstanceNormalization()(conv1)
    conv1 = keras.layers.PReLU(shared_axes=[1])(conv1)
    conv1 = keras.layers.Dropout(rate=0.2)(conv1)
    conv1 = keras.layers.MaxPooling1D(pool_size=2)(conv1)
    conv2 = keras.layers.Conv1D(filters=256, kernel_size=11, strides=1, padding='same')(conv1)
    conv2 = InstanceNormalization()(conv2)
    conv2 = keras.layers.PReLU(shared_axes=[1])(conv2)
    conv2 = keras.layers.Dropout(rate=0.2)(conv2)
    conv2 = keras.layers.MaxPooling1D(pool_size=2)(conv2)
    conv3 = keras.layers.Conv1D(filters=512, kernel_size=21, strides=1, padding='same')(conv2)
    conv3 = InstanceNormalization()(conv3)
    conv3 = keras.layers.PReLU(shared_axes=[1])(conv3)
    conv3 = keras.layers.Dropout(rate=0.2)(conv3)
    attention_data = keras.layers.Lambda(lambda x: x[:, :, :256])(conv3)
    attention_softmax = keras.layers.Lambda(lambda x: x[:, :, 256:])(conv3)
    attention_softmax = keras.layers.Softmax()(attention_softmax)
    multiply_layer = keras.layers.Multiply()([attention_softmax, attention_data])
    dense_layer = keras.layers.Dense(units=256, activation='sigmoid')(multiply_layer)
    dense_layer = InstanceNormalization()(dense_layer)
    flatten_layer = keras.layers.Flatten()(dense_layer)
    output_layer = keras.layers.Dense(units=nb_classes, activation='softmax')(flatten_layer)

    model = keras.models.Model(inputs=input_layer, outputs=output_layer)

    model.compile(loss='categorical_crossentropy', optimizer=keras.optimizers.Adam(0.00001), metrics=['accuracy'])

    return model, []


# Additional TSC models implemented here for v2
def dl4tsc_inception(input_layer, nb_classes) -> Tuple["keras.Model", List[Callable]]:  
    nb_filters = 32
    use_residual = True
    use_bottleneck = True
    depth = 6
    kernel_size = 41 - 1  # follow legacy impl offset
    bottleneck_size = 32

    def inception_module(x, stride=1, activation='linear'):
        if use_bottleneck and int(x.shape[-1]) > bottleneck_size:
            input_inception = keras.layers.Conv1D(
                filters=bottleneck_size, kernel_size=1, padding='same', activation=activation, use_bias=False
            )(x)
        else:
            input_inception = x

        kernel_size_s = [kernel_size // (2 ** i) for i in range(3)]
        conv_list = [
            keras.layers.Conv1D(filters=nb_filters, kernel_size=ks, strides=stride,
                                 padding='same', activation=activation, use_bias=False)(input_inception)
            for ks in kernel_size_s
        ]

        max_pool_1 = keras.layers.MaxPool1D(pool_size=3, strides=stride, padding='same')(x)
        conv_6 = keras.layers.Conv1D(filters=nb_filters, kernel_size=1,
                                     padding='same', activation=activation, use_bias=False)(max_pool_1)
        conv_list.append(conv_6)

        x_out = keras.layers.Concatenate(axis=2)(conv_list)
        x_out = keras.layers.BatchNormalization()(x_out)
        x_out = keras.layers.Activation('relu')(x_out)
        return x_out

    def shortcut_layer(x_in, x_out):
        shortcut_y = keras.layers.Conv1D(filters=int(x_out.shape[-1]), kernel_size=1,
                                         padding='same', use_bias=False)(x_in)
        shortcut_y = keras.layers.BatchNormalization()(shortcut_y)
        x = keras.layers.Add()([shortcut_y, x_out])
        x = keras.layers.Activation('relu')(x)
        return x

    x = input_layer
    input_res = input_layer
    for d in range(depth):
        x = inception_module(x)
        if use_residual and d % 3 == 2:
            x = shortcut_layer(input_res, x)
            input_res = x

    gap_layer = keras.layers.GlobalAveragePooling1D()(x)
    output_layer = keras.layers.Dense(nb_classes, activation='softmax')(gap_layer)
    model = keras.models.Model(inputs=input_layer, outputs=output_layer)

    model.compile(loss='categorical_crossentropy', optimizer=keras.optimizers.Adam(0.001), metrics=['accuracy'])
    reduce_lr = keras.callbacks.ReduceLROnPlateau(monitor='loss', factor=0.5, patience=50, min_lr=0.0001)
    return model, [reduce_lr]


def dl4tsc_tlenet(input_layer, nb_classes) -> Tuple["keras.Model", List[Callable]]:  
    conv_1 = keras.layers.Conv1D(filters=5, kernel_size=5, activation='relu', padding='same')(input_layer)
    conv_1 = keras.layers.MaxPool1D(pool_size=2)(conv_1)

    conv_2 = keras.layers.Conv1D(filters=20, kernel_size=5, activation='relu', padding='same')(conv_1)
    conv_2 = keras.layers.MaxPool1D(pool_size=4)(conv_2)

    flatten_layer = keras.layers.Flatten()(conv_2)
    fully_connected_layer = keras.layers.Dense(500, activation='relu')(flatten_layer)

    output_layer = keras.layers.Dense(nb_classes, activation='softmax')(fully_connected_layer)
    model = keras.models.Model(inputs=input_layer, outputs=output_layer)

    model.compile(optimizer=keras.optimizers.Adam(learning_rate=0.01, decay=0.005),
                  loss='categorical_crossentropy', metrics=['accuracy'])
    return model, []


def dl4tsc_mcdcnn(input_layer, nb_classes) -> Tuple["keras.Model", List[Callable]]:  
    n_t = int(input_layer.shape[1])
    n_vars = int(input_layer.shape[2]) if len(input_layer.shape) > 2 else 1

    padding = 'valid'
    if n_t < 60:
        padding = 'same'

    input_layers = []
    conv2_layers = []
    for _ in range(n_vars):
        in_i = keras.layers.Input((n_t, 1))
        input_layers.append(in_i)
        c1 = keras.layers.Conv1D(filters=8, kernel_size=5, activation='relu', padding=padding)(in_i)
        c1 = keras.layers.MaxPooling1D(pool_size=2)(c1)
        c2 = keras.layers.Conv1D(filters=8, kernel_size=5, activation='relu', padding=padding)(c1)
        c2 = keras.layers.MaxPooling1D(pool_size=2)(c2)
        c2 = keras.layers.Flatten()(c2)
        conv2_layers.append(c2)

    if n_vars == 1:
        concat_layer = conv2_layers[0]
    else:
        concat_layer = keras.layers.Concatenate(axis=-1)(conv2_layers)

    fully_connected = keras.layers.Dense(units=732, activation='relu')(concat_layer)
    output_layer = keras.layers.Dense(nb_classes, activation='softmax')(fully_connected)
    model = keras.models.Model(inputs=input_layers, outputs=output_layer)

    # Match legacy optimizer config
    model.compile(loss='categorical_crossentropy',
                  optimizer=keras.optimizers.SGD(learning_rate=0.01, momentum=0.9, decay=0.0005),
                  metrics=['accuracy'])
    return model, []


def dl4tsc_mcnn(input_layer, nb_classes) -> Tuple["keras.Model", List[Callable]]:  
    # Derive number of variables from input layer shape (T, V)
    n_t = int(input_layer.shape[1])
    n_vars = int(input_layer.shape[2]) if len(input_layer.shape) > 2 else 1

    pool_factor = 2
    kernel_size = max(3, min(11, n_t // 10))

    stage_1_layers = []
    input_layers = []
    for _ in range(n_vars):
        in_i = keras.layers.Input((n_t, 1))
        input_layers.append(in_i)
        conv_layer = keras.layers.Conv1D(filters=256, kernel_size=kernel_size, padding='same', activation='sigmoid')(in_i)
        pool_size = max(1, int(int(conv_layer.shape[1]) / pool_factor))
        max_layer = keras.layers.MaxPooling1D(pool_size=pool_size)(conv_layer)
        stage_1_layers.append(max_layer)

    concat_layer = stage_1_layers[0] if n_vars == 1 else keras.layers.Concatenate(axis=-1)(stage_1_layers)
    kernel_size_2 = int(min(kernel_size, int(concat_layer.shape[1])))
    full_conv = keras.layers.Conv1D(filters=256, kernel_size=kernel_size_2, padding='same', activation='sigmoid')(concat_layer)
    pool_size_2 = max(1, int(int(full_conv.shape[1]) / pool_factor))
    full_max = keras.layers.MaxPooling1D(pool_size=pool_size_2)(full_conv)
    full_max = keras.layers.Flatten()(full_max)
    fully_connected = keras.layers.Dense(units=256, activation='sigmoid')(full_max)

    output_layer = keras.layers.Dense(nb_classes, activation='softmax')(fully_connected)
    model = keras.models.Model(inputs=input_layers, outputs=output_layer)
    # Follow model_builder’s config for MCNN
    model.compile(loss='categorical_crossentropy', optimizer=keras.optimizers.Adam(0.1), metrics=['accuracy'])
    return model, []


def _legacy_create_directory(path: str):
    """Create a directory if it does not exist (legacy helper).

    Mirrors the minimal behavior of the legacy `create_directory` util used in
    the original TSC codebase. Safe to call repeatedly.
    """
    os.makedirs(path, exist_ok=True)


class dl4tsc_twiesn:
    def __init__(self, output_dir: str = "TWIESN_model_out"):
        # This class is a lightweight, self-contained port of the legacy TWIESN
        # model for v2. It intentionally avoids inheriting from legacy bases.
        self.output_dir = output_dir
        # hyperparameters
        first_config = {'N_x': 250, 'connect': 0.5, 'scaleW_in': 1.0, 'lamda': 0.0}
        second_config = {'N_x': 250, 'connect': 0.5, 'scaleW_in': 2.0, 'lamda': 0.05}
        third_config = {'N_x': 500, 'connect': 0.1, 'scaleW_in': 2.0, 'lamda': 0.05}
        fourth_config = {'N_x': 800, 'connect': 0.1, 'scaleW_in': 2.0, 'lamda': 0.05}
        self.configs = [first_config, second_config, third_config, fourth_config]
        self.rho_s = [0.55, 0.9, 2.0, 5.0]
        self.alpha = 0.1  # leaky rate
        self.y_pred = None
        self.y_pred_reshaped = None

    def persist_model_trainables(self, path_like: str = 'saved_trainable_values', **kwargs):
        raise NotImplementedError()

    def restore_model_trainables(self, path_like, dataset = None, **kwargs):
        raise NotImplementedError()

    def init_matrices(self):
        self.W_in = (2.0 * np.random.rand(self.N_x, self.num_dim) - 1.0) / (2.0 * self.scaleW_in)

        converged = False
        i = 0
        while not converged:
            i += 1
            self.W = sparse.rand(self.N_x, self.N_x, density=self.connect).todense()
            self.W[np.where(self.W > 0)] -= 0.5
            try:
                eig, _ = slinalg.eigs(self.W, k=1, which='LM')
                converged = True
            except Exception:
                print('not converged ', i)
                continue
        self.W /= np.abs(eig) / self.rho

    def compute_state_matrix(self, x_in):
        n = x_in.shape[0]
        X_t = np.zeros((n, self.T, self.N_x), dtype=np.float64)
        X_t_1 = np.zeros((n, self.N_x), dtype=np.float64)
        for t in range(self.T):
            curr_in = x_in[:, t, :]
            curr_state = np.tanh(self.W_in.dot(curr_in.T) + self.W.dot(X_t_1.T)).T
            curr_state = (1 - self.alpha) * X_t_1 + self.alpha * curr_state
            X_t_1 = curr_state
            X_t[:, t, :] = curr_state
        return X_t

    def reshape_prediction(self, y_pred, num_instances, length_series):
        new_y_pred = y_pred.reshape(num_instances, length_series, y_pred.shape[-1])
        new_y_pred = np.average(new_y_pred, axis=1)
        new_y_pred = np.argmax(new_y_pred, axis=1)
        return new_y_pred

    def train(self):
        start_time = time.time()
        self.init_matrices()
        state_matrix = self.compute_state_matrix(self.x_train)
        new_x_train = np.concatenate((self.x_train, state_matrix), axis=2).reshape(
            self.N * self.T, self.num_dim + self.N_x)
        state_matrix = None
        gc.collect()
        new_labels = np.repeat(self.y_train, self.T, axis=0)
        ridge_classifier = Ridge(alpha=self.lamda)
        ridge_classifier.fit(new_x_train, new_labels)

        state_matrix = self.compute_state_matrix(self.x_val)
        new_x_val = np.concatenate((self.x_val, state_matrix), axis=2).reshape(
            self.x_val.shape[0] * self.T, self.num_dim + self.N_x)
        y_pred_val = ridge_classifier.predict(new_x_val)
        y_pred_val = self.reshape_prediction(y_pred_val, self.x_val.shape[0], self.T)
        df_val_metrics = self.calculate_metrics(np.argmax(self.y_val, axis=1), y_pred_val, 0.0)
        train_acc = df_val_metrics['accuracy'][0]

        state_matrix = self.compute_state_matrix(self.x_test)
        new_x_test = np.concatenate((self.x_test, state_matrix), axis=2).reshape(self.x_test.shape[0] * self.T,
                                                                                 self.num_dim + self.N_x)
        state_matrix = None
        gc.collect()
        self.y_pred = y_pred = ridge_classifier.predict(new_x_test)
        self.y_pred_reshaped = y_pred_reshaped = self.reshape_prediction(y_pred, self.x_test.shape[0], self.T)

        duration = time.time() - start_time
        df_metrics = self.calculate_metrics(self.y_true, y_pred_reshaped, duration)
        self.W_out = ridge_classifier.coef_
        gc.collect()
        np.savetxt(os.path.join(self.output_dir, 'W_in.txt'), self.W_in)
        np.savetxt(os.path.join(self.output_dir, 'W.txt'), self.W)
        np.savetxt(os.path.join(self.output_dir, 'W_out.txt'), self.W_out)
        df_metrics.to_csv(os.path.join(self.output_dir, 'df_metrics.csv'), index=False)
        return df_metrics, train_acc, ridge_classifier

    def convert_dataset_to_legacy_values(self, dataset) -> (
        np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.ndarray, int, int
    ):
        """Convert v2 Dataset into legacy tuple expected by dl4tsc implementations.

        Returns (x_train, x_val, y_train, y_val, y_true, input_shape, nb_classes).
        """
        # Import locally to avoid hard dependency at module import time
        from useckit.v2.dataset import Dataset as V2Dataset

        if not isinstance(dataset, V2Dataset):
            # Be permissive: assume a duck-typed object with the same interface
            pass

        x_train, x_val = dataset.trainset_data, dataset.validationset_data
        y_train, y_val, _, _ = dataset.view_one_hot_encoded_labels()

        y_true = np.argmax(y_val, axis=1)
        nb_classes = y_train.shape[1] if y_train is not None else int(np.max(y_true) + 1)
        input_shape = x_train.shape[1:]

        assert isinstance(nb_classes, int)
        assert x_train.shape[0] == y_train.shape[0]
        assert x_val.shape[0] == y_val.shape[0]

        return x_train, x_val, y_train, y_val, y_true, input_shape, nb_classes

    def fit(self, dataset, *args, **kwargs):
        self.x_train, self.x_val, self.y_train, self.y_val, y_true, input_shape, nb_classes = self.convert_dataset_to_legacy_values(dataset)

        best_train_acc = -1

        self.num_dim = self.x_train.shape[2]
        self.T = self.x_train.shape[1]
        self.x_test = self.x_val
        self.y_true = y_true
        self.y_test = self.y_val

        self.N = self.x_train.shape[0]

        if self.x_train.shape[0] > 1000 or self.x_test.shape[0] > 1000:
            for config in self.configs:
                config['N_x'] = 100
            self.configs = [self.configs[0], self.configs[1], self.configs[2]]

        output_directory_root = self.output_dir
        for idx_config in range(len(self.configs)):
            for rho in self.rho_s:
                self.rho = rho
                self.N_x = self.configs[idx_config]['N_x']
                self.connect = self.configs[idx_config]['connect']
                self.scaleW_in = self.configs[idx_config]['scaleW_in']
                self.lamda = self.configs[idx_config]['lamda']
                self.output_dir = os.path.join(output_directory_root,
                                               'hyper_param_search',
                                               'config_' + str(idx_config),
                                               'rho_' + str(rho))
                _legacy_create_directory(self.output_dir)
                df_metrics, train_acc, clf = self.train()

                if best_train_acc < train_acc:
                    best_train_acc = train_acc
                    self.y_pred_best = self.y_pred
                    self.y_pred_reshaped_best = self.y_pred_reshaped
                    df_metrics.to_csv(os.path.join(output_directory_root, 'df_metrics.csv'), index=False)
                    np.savetxt(os.path.join(output_directory_root, 'W_in.txt'), self.W_in)
                    np.savetxt(os.path.join(output_directory_root, 'W.txt'), self.W)
                    np.savetxt(os.path.join(output_directory_root, 'W_out.txt'), self.W_out)
                    with open(os.path.join(output_directory_root, 'ridge.pkl'), 'wb') as file:
                        pickle.dump(clf, file)
                    with open(os.path.join(output_directory_root, 'T.pkl'), 'wb') as file:
                        pickle.dump(self.T, file)
                    with open(os.path.join(output_directory_root, 'N_x.pkl'), 'wb') as file:
                        pickle.dump(self.N_x, file)
                    with open(os.path.join(output_directory_root, 'alpha.pkl'), 'wb') as file:
                        pickle.dump(self.alpha, file)

                gc.collect()

    def predict(self, x_test, y_true, x_train, y_train, y_test, return_df_metrics=True):
        from pathlib import Path
        modelroot = Path(self.output_dir).parent.parent.parent
        with open(modelroot.joinpath("ridge.pkl"), 'rb') as file:
            ridge_classifier = pickle.load(file)
        with open(modelroot.joinpath("T.pkl"), 'rb') as file:
            self.T = pickle.load(file)
        with open(modelroot.joinpath("N_x.pkl"), 'rb') as file:
            self.N_x = pickle.load(file)
        with open(modelroot.joinpath("alpha.pkl"), 'rb') as file:
            self.alpha = pickle.load(file)
        self.W_in = np.loadtxt(modelroot.joinpath('W_in.txt'))
        self.W = np.loadtxt(modelroot.joinpath('W.txt'))
        self.W_out = np.loadtxt(modelroot.joinpath('W_out.txt'))

        state_matrix = self.compute_state_matrix(x_test)
        new_x_test = np.concatenate((x_test, state_matrix), axis=2).reshape(x_test.shape[0] * self.T,
                                                                            self.num_dim + self.N_x)
        state_matrix = None
        gc.collect()
        y_pred = ridge_classifier.predict(new_x_test)
        y_pred_reshaped = self.reshape_prediction(y_pred, x_test.shape[0], self.T)

        return y_pred_reshaped

    def calculate_metrics(self, y_true, y_pred, duration, y_true_val=None, y_pred_val=None):
        import pandas as pd
        from sklearn.metrics import precision_score, accuracy_score, recall_score

        res = pd.DataFrame(data=np.zeros((1, 4), dtype=np.float64), index=[0],
                           columns=['precision', 'accuracy', 'recall', 'duration'])

        res['precision'] = precision_score(y_true, y_pred, average='macro')
        res['accuracy'] = accuracy_score(y_true, y_pred)

        if not y_true_val is None:
            # this is useful when transfer learning is used with cross validation
            res['accuracy_val'] = accuracy_score(y_true_val, y_pred_val)

        res['recall'] = recall_score(y_true, y_pred, average='macro')
        res['duration'] = duration
        return res



MODEL_DESCRIPTIONS = {
    'CNN_VALID': dl4tsc_cnn_padding_valid,
    'CNN_SAME': dl4tsc_cnn_padding_same,
    'FCN': dl4tsc_fcn,
    'MLP': dl4tsc_mlp,
    'RESNET': dl4tsc_resnet,
    'ENCODER': dl4tsc_encoder,
    'INCEPTION': dl4tsc_inception,
    'TLENET': dl4tsc_tlenet,
    'MCDCNN': dl4tsc_mcdcnn,
    'MCNN': dl4tsc_mcnn,
    # Non-Keras
    'TWIESN': dl4tsc_twiesn,
}



import os
from typing import Callable

import tensorflow as tf
from tensorflow.keras import metrics
from tensorflow.keras import Model

from useckit.paradigms.distance_learning.prediction_models.contrastive_loss.contrastive_model_description import \
    tower_mlp as tower_definition
from useckit.paradigms.distance_learning.prediction_models.offline_triplet_loss.triplet_distance_layer import \
    TrippleTowerDistanceLayer


class TrippleTowerModel(Model):
    """The Siamese Network model with a custom training and testing loops.

    Computes the triplet loss using the three embeddings produced by the
    Siamese Network.

    The triplet loss is defined as:
       L(A, P, N) = max(‖f(A) - f(P)‖² - ‖f(A) - f(N)‖² + margin, 0)
    """

    def __init__(self, tower_input_shape=None,
                 tower_model_description: Callable = tower_definition,
                 output_dir=None,
                 margin=0.5):
        super().__init__()
        self.tower_model_description = tower_model_description
        self.output_dir = output_dir

        if tower_input_shape is None:
            raise ValueError("Parameter `tower_input_shape` must be provided.")

        self.input_layer = tf.keras.layers.Input(tower_input_shape)

        self.tower_1 = self.tower_model_description(self.input_layer)  # anchor
        self.tower_2 = self.tower_model_description(self.input_layer)  # positive
        self.tower_3 = self.tower_model_description(self.input_layer)  # negative

        self.merge = TrippleTowerDistanceLayer()

        model_checkpoint = tf.keras.callbacks.ModelCheckpoint(
            filepath=os.path.join(self.output_dir, 'best_model.weights.h5'),
            monitor='val_loss',
            save_best_only=True,
            save_weights_only=True)
        self.callbacks = [model_checkpoint]

        self.compile(optimizer=tf.keras.optimizers.Adam(0.0001))

        self.margin = margin
        self.loss_tracker = metrics.Mean(name="loss")

    def call(self, inputs):
        x1 = self.tower_1(inputs[0])
        x2 = self.tower_2(inputs[1])
        x3 = self.tower_3(inputs[2])
        return self.merge(x1, x2, x3)

    def train_step(self, data):
        # GradientTape is a context manager that records every operation that
        # you do inside. We are using it here to compute the loss so we can get
        # the gradients and apply them using the optimizer specified in
        # `compile()`.
        with tf.GradientTape() as tape:
            loss = self._compute_loss(data)

        # Storing the gradients of the loss function with respect to the
        # weights/parameters.
        gradients = tape.gradient(loss, self.trainable_weights)

        # Applying the gradients on the model using the specified optimizer
        self.optimizer.apply_gradients(
            zip(gradients, self.trainable_weights)
        )

        # Let's update and return the training loss metric.
        self.loss_tracker.update_state(loss)
        return {"loss": self.loss_tracker.result()}

    def test_step(self, data):
        loss = self._compute_loss(data)

        # Let's update and return the loss metric.
        self.loss_tracker.update_state(loss)
        return {"loss": self.loss_tracker.result()}

    def _compute_loss(self, data):
        # The output of the network is a tuple containing the distances
        # between the anchor and the positive example, and the anchor and
        # the negative example.
        ap_distance, an_distance = self(data[0])

        # Computing the Triplet Loss by subtracting both distances and
        # making sure we don't get a negative value.
        loss = ap_distance - an_distance
        loss = tf.maximum(loss + self.margin, 0.0)
        return loss

    @property
    def metrics(self):
        # We need to list our metrics here so the `reset_states()` can be
        # called automatically.
        return [self.loss_tracker]



from useckit.paradigms.time_series_classification.tsc_paradigm import TSCParadigm
from useckit.paradigms.anomaly_detection.anomaly_paradigm import AnomalyParadigm
from useckit.paradigms.distance_learning.distance_paradigm import DistanceMetricParadigm
from useckit.util.dataset import Dataset


def set_seed(seed: int = 42) -> None:
    """Sets RNG seed to python, numpy, tensorflow, and environment to the provided number."""
    from useckit.paradigms._paradigm_base import ParadigmBase
    ParadigmBase.set_seed(seed)



import os
import pickle

import numpy as np

from useckit import Dataset
from useckit.paradigms.anomaly_detection.prediction_models.scikit_model_descriptions import ScikitBaseDescription
from useckit.paradigms.binary_verification.prediction_models.scikit_model_descriptions import ScikitClassif
from useckit.paradigms.binary_verification.prediction_models.verification_prediction_model_base import \
    VerificationBasePredictionModel


class VerificationSkLearnPredictionModel(VerificationBasePredictionModel):
    def __init__(self,
                 scikit_binary_classif: ScikitBaseDescription = ScikitClassif(),
                 output_dir: str = "scikit_pred_model_out",
                 verbose=False):
        super().__init__(output_dir=output_dir, verbose=verbose)
        self.build_description = scikit_binary_classif
        self.classifs = []
        self.labels_per_classif = []

    def persist_model_trainables(self, path_like: str = 'saved_trainable_values', **kwargs):
        classifs_file = os.path.join(self.output_dir, path_like)
        labels_per_classif_file = os.path.join(os.path.dirname(os.path.join(self.output_dir, path_like)),
                                                "labels_per_classif.persisted")
        with open(classifs_file, "wb") as file:
            pickle.dump(self.classifs, file)
        with open(labels_per_classif_file, "wb") as file:
            pickle.dump(self.labels_per_classif, file)

    def restore_model_trainables(self, path_like: str, **kwargs):
        classifs_file = os.path.join(self.output_dir, path_like)
        labels_per_classif_file = os.path.join(os.path.dirname(os.path.join(self.output_dir, path_like)),
                                                "labels_per_classif.persisted")
        with open(classifs_file, "rb") as file:
            self.classifs = pickle.load(file)
        with open(labels_per_classif_file, "rb") as file:
            self.labels_per_classif = pickle.load(file)

    def build_model(self, classes, **kwargs) -> dict:
        """Returns a dict of built models. The key of the dict is:
        * in OVR case: the class that the model was built for, or
        * in OVO case: a defaultdict where the first layer is class1 and the second layer is class2."""
        strat: str = kwargs["multiclass_classification_strategy"]

        if strat == "ovr":
            result = dict()
            for c in classes:
                assert c not in result.keys(), 'Error 1: key already exists.'
                # Assuming 'build_description.build_model()' method builds a model for the specific class
                result[c] = self.build_description.build_model()
        elif strat == "ovo":
            from collections import defaultdict
            result = defaultdict(dict)

            for i, c1 in enumerate(classes):
                for j, c2 in enumerate(classes):
                    if i >= j:  # Skip already considered or same class pairs
                        continue
                    result[c2][c1] = self.build_description.build_model()
        return result

    """def fit(self, dataset: Dataset, **kwargs):
        x_train, y_train = dataset.trainset_data, dataset.trainset_labels
        x_train = np.array([x.flatten() for x in x_train])
        unique_train_labels = dataset.get_unique_labels()
        assert len(unique_train_labels) > 0, "Dataset must have one or more classes to train one."
        classifs_built_dict: dict = self.build_model(unique_train_labels, **kwargs)

        # fit the different classifs
        for key in classifs_built_dict.keys():
            # find the elements in the training data that belong to class `i`
            if kwargs['multiclass_classification_strategy'] == "ovr":
                data_select = y_train == key
                y_train_clf = data_select.astype(int)  # cast boolean mask to int (0=False and 1=True)
                classifs_built_dict[key].fit(x_train, y_train_clf)
            elif kwargs['multiclass_classification_strategy'] == "ovo":
                for subkey in classifs_built_dict[key].keys():
                    left_model_key = key
                    right_model_key = subkey
                    data_select_left_accept = (y_train == left_model_key)
                    data_select_right_reject = (y_train == right_model_key)
                    data_to_accept = x_train[data_select_left_accept]
                    data_to_reject = x_train[data_select_right_reject]
                    x_train_clf = np.concatenate([data_to_accept, data_to_reject])
                    y_train_clf = [1] * len(data_to_accept) + [0] * len(data_to_reject)
                    classifs_built_dict[key][subkey].fit(x_train_clf, y_train_clf)

        self.classifs = classifs_built_dict"""

    def fit(self, dataset: Dataset, enable_parallel_fitting: bool = True, **kwargs):
        from joblib import Parallel, delayed

        x_train, y_train = dataset.trainset_data, dataset.trainset_labels
        x_train = np.array([x.flatten() for x in x_train])
        unique_train_labels = dataset.get_unique_labels()
        assert len(unique_train_labels) > 0, "Dataset must have one or more classes to train one."
        classifs_built_dict: dict = self.build_model(unique_train_labels, **kwargs)

        def fit_ovr_classifier(key, classifier):
            data_select = y_train == key
            y_train_clf = data_select.astype(int)  # cast boolean mask to int (0=False and 1=True)
            classifier.fit(x_train, y_train_clf)
            return key, classifier

        def fit_ovo_classifier(key, subkey, classifier):
            left_model_key = key
            right_model_key = subkey
            data_select_left_accept = (y_train == left_model_key)
            data_select_right_reject = (y_train == right_model_key)
            data_to_accept = x_train[data_select_left_accept]
            data_to_reject = x_train[data_select_right_reject]
            x_train_clf = np.concatenate([data_to_accept, data_to_reject])
            y_train_clf = [1] * len(data_to_accept) + [0] * len(data_to_reject)
            classifier.fit(x_train_clf, y_train_clf)
            return key, subkey, classifier

        if enable_parallel_fitting:
            if kwargs['multiclass_classification_strategy'] == "ovr":
                results = Parallel(n_jobs=-1)(
                    delayed(fit_ovr_classifier)(key, classifs_built_dict[key]) for key in classifs_built_dict.keys()
                )
                for key, classifier in results:
                    classifs_built_dict[key] = classifier

            elif kwargs['multiclass_classification_strategy'] == "ovo":
                results = Parallel(n_jobs=-1)(
                    delayed(fit_ovo_classifier)(key, subkey, classifs_built_dict[key][subkey])
                    for key in classifs_built_dict.keys()
                    for subkey in classifs_built_dict[key].keys()
                )
                for key, subkey, classifier in results:
                    classifs_built_dict[key][subkey] = classifier
        else:
            # fit the different classifs sequentially
            for key in classifs_built_dict.keys():
                if kwargs['multiclass_classification_strategy'] == "ovr":
                    _, classifs_built_dict[key] = fit_ovr_classifier(key, classifs_built_dict[key])
                elif kwargs['multiclass_classification_strategy'] == "ovo":
                    for subkey in classifs_built_dict[key].keys():
                        _, _, classifs_built_dict[key][subkey] = fit_ovo_classifier(key, subkey,
                                                                                    classifs_built_dict[key][subkey])

        self.classifs = classifs_built_dict

    def predict(self, x_test):
        x_test = x_test.reshape((x_test.shape[0], np.prod(x_test.shape[1:])))
        result_pred = []
        result_id = []
        for i, classif in zip(self.labels_per_classif, self.classifs):
            preds = classif.predict(x_test)
            result_pred.append(preds)
            result_id.append(i)
        return np.array(result_pred), np.array(result_id)



from .evaluation_methods.identification import IdentificationOnly
from .evaluation_methods.tsc_evaluation_method_base import TSCBaseEvaluationMethod
from .prediction_models.classification_keras_prediction_model import ClassificationKerasPredictionModel
from .prediction_models.tsc_prediction_model_base import TSCBasePredictionModel
from .._paradigm_base import ParadigmBase, PredictionModelBase
from ...util.dataset import Dataset


class TSCParadigm(ParadigmBase):

    def __init__(self,
                 prediction_model: TSCBasePredictionModel = None,
                 evaluation_methods: [TSCBaseEvaluationMethod] = None,
                 seed=42,
                 experiment_description_name=None,
                 disable_normalization_check=False,
                 class_weights=None,
                 **base_kwargs
                 ):
        # This looks really stupid, but if the default value is initiated in the method head, not in the call to the
        # super constructor as below, python will only instantiate one object for all calls to this method, hence
        # causing different paradigms to share the same instances of prediction model or evaluation method.
        # For more information see:
        # https://florimond.dev/en/posts/2018/08/python-mutable-defaults-are-the-source-of-all-evil/
        super().__init__(
            prediction_model=ClassificationKerasPredictionModel() if prediction_model is None else prediction_model,
            evaluation_methods=
            [IdentificationOnly()] if evaluation_methods is None else evaluation_methods,
            seed=seed,
            experiment_description_name=experiment_description_name,
            **base_kwargs)
        self.disable_normalization_check = disable_normalization_check
        self.prediction_model = prediction_model

        if class_weights is not None:
            assert isinstance(class_weights, dict), 'class_weights must be a dict. See: https://bit.ly/3ljOI9c'
        self.class_weights = class_weights

    def restore_prediction_model_trainables(self,
                                            saved_weights_path_like: str,
                                            dataset: Dataset) -> PredictionModelBase:
        if dataset is None:
            raise ValueError(
                "The dataset argument may not be None, as the method uses it to infer the size of the " +
                "restored model's first layer. " +
                "Note that the total amount of unique labels within dataset needs to match the total amount of " +
                "labels in the dataset that the model was trained on.")
        self.prediction_model.restore_model_trainables(saved_weights_path_like, dataset=dataset)
        return self.prediction_model



import os
import pickle

import numpy as np

from .anomaly_prediction_model_base import AnomalyBasePredictionModel
from .scikit_model_descriptions import ScikitBaseDescription, ScikitRegressor
from ....util.dataset import Dataset


class ScikitAnomalyPredictionModel(AnomalyBasePredictionModel):

    def __init__(self,
                 scikit_regressor: ScikitBaseDescription = ScikitRegressor(),
                 output_dir: str = "scikit_pred_model_out",
                 verbose=False):
        super().__init__(output_dir=output_dir, verbose=verbose)
        self.build_description = scikit_regressor
        self.regressors = []
        self.label_per_regressor = []

    def persist_model_trainables(self, path_like: str = 'saved_trainable_values', **kwargs):
        regressors_file = os.path.join(self.output_dir, path_like)
        label_per_regressor_file = os.path.join(os.path.dirname(os.path.join(self.output_dir, path_like)),
                                                "labels_per_regressor.persisted")
        with open(regressors_file, "wb") as file:
            pickle.dump(self.regressors, file)
        with open(label_per_regressor_file, "wb") as file:
            pickle.dump(self.label_per_regressor, file)

    def restore_model_trainables(self, path_like: str, **kwargs):
        regressors_file = os.path.join(self.output_dir, path_like)
        label_per_regressor_file = os.path.join(os.path.dirname(os.path.join(self.output_dir, path_like)),
                                                "labels_per_regressor.persisted")
        with open(regressors_file, "rb") as file:
            self.regressors = pickle.load(file)
        with open(label_per_regressor_file, "rb") as file:
            self.label_per_regressor = pickle.load(file)

    def build_model(self, nb_classes):
        result = []
        for i in range(nb_classes):
            result.append(self.build_description.build_model())
        return result

    def fit(self, dataset: Dataset):
        x_train, y_train = dataset.trainset_data, dataset.trainset_labels
        x_train = x_train.reshape((x_train.shape[0], np.prod(x_train.shape[1:])))
        nb_classes = dataset.train_classes()
        assert nb_classes > 0
        regressors_built = self.build_model(nb_classes)
        for i, regressor in zip(np.unique(y_train), regressors_built):
            # find the elements in the training data that belong to class `i`
            belonging_indexes = y_train == i

            true_anomaly_scores = np.ones((len(x_train),), dtype='float16')
            true_anomaly_scores[belonging_indexes] = 0

            regressor.fit(x_train, true_anomaly_scores)
            self.label_per_regressor.append(i)
        self.regressors = regressors_built

    def predict(self, x_test):
        x_test = x_test.reshape((x_test.shape[0], np.prod(x_test.shape[1:])))
        result_pred = []
        result_id = []
        for i, regressor in zip(self.label_per_regressor, self.regressors):
            preds = regressor.predict(x_test)
            result_pred.append(preds)
            result_id.append(i)
        return np.array(result_pred), np.array(result_id)
